from __future__ import annotations

import asyncio
import logging
import uuid
from typing import Final, Optional, cast
from uuid import UUID

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.bgtask.reporter import ProgressReporter
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.data.artifact.types import (
    ArtifactRegistryType,
    ArtifactRevisionDownloadProgress,
    CombinedDownloadProgress,
)
from ai.backend.common.data.storage.registries.types import ModelTarget
from ai.backend.common.data.storage.types import ArtifactStorageType
from ai.backend.common.dto.storage.request import (
    DeleteObjectReq,
    HuggingFaceGetCommitHashReqPathParam,
    HuggingFaceGetCommitHashReqQueryParam,
    HuggingFaceImportModelsReq,
    ReservoirImportModelsReq,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.client.artifact_registry.reservoir_client import ReservoirRegistryClient
from ai.backend.manager.clients.storage_proxy.session_manager import StorageSessionManager
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.config.unified import (
    ReservoirConfig,
    ReservoirObjectStorageConfig,
    ReservoirVFSStorageConfig,
)
from ai.backend.manager.data.artifact.types import (
    ArtifactRevisionData,
    ArtifactRevisionReadme,
    ArtifactStatus,
    ArtifactType,
)
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryData
from ai.backend.manager.dto.request import DelegateImportArtifactsReq
from ai.backend.manager.errors.artifact import (
    ArtifactDeletionBadRequestError,
    ArtifactDeletionError,
    ArtifactImportBadRequestError,
    RemoteReservoirArtifactImportError,
)
from ai.backend.manager.errors.artifact_registry import (
    InvalidArtifactRegistryTypeError,
)
from ai.backend.manager.errors.common import ServerMisconfiguredError
from ai.backend.manager.errors.storage import UnsupportedStorageTypeError
from ai.backend.manager.repositories.artifact.repository import ArtifactRepository
from ai.backend.manager.repositories.artifact_registry.repository import ArtifactRegistryRepository
from ai.backend.manager.repositories.huggingface_registry.repository import HuggingFaceRepository
from ai.backend.manager.repositories.object_storage.repository import ObjectStorageRepository
from ai.backend.manager.repositories.reservoir_registry.repository import (
    ReservoirRegistryRepository,
)
from ai.backend.manager.repositories.storage_namespace.repository import StorageNamespaceRepository
from ai.backend.manager.repositories.vfs_storage.repository import VFSStorageRepository
from ai.backend.manager.services.artifact_revision.actions.approve import (
    ApproveArtifactRevisionAction,
    ApproveArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.associate_with_storage import (
    AssociateWithStorageAction,
    AssociateWithStorageActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.cancel_import import (
    CancelImportAction,
    CancelImportActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.cleanup import (
    CleanupArtifactRevisionAction,
    CleanupArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.delegate_import_revision_batch import (
    DelegateImportArtifactRevisionBatchAction,
    DelegateImportArtifactRevisionBatchActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.disassociate_with_storage import (
    DisassociateWithStorageAction,
    DisassociateWithStorageActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get import (
    GetArtifactRevisionAction,
    GetArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get_download_progress import (
    GetDownloadProgressAction,
    GetDownloadProgressActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get_readme import (
    GetArtifactRevisionReadmeAction,
    GetArtifactRevisionReadmeActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.import_revision import (
    ImportArtifactRevisionAction,
    ImportArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.list import (
    ListArtifactRevisionsAction,
    ListArtifactRevisionsActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.reject import (
    RejectArtifactRevisionAction,
    RejectArtifactRevisionActionResult,
)

_REMOTE_ARTIFACT_STATUS_POLL_INTERVAL: Final[int] = 30  # seconds
_REMOTE_ARTIFACT_MAX_WAIT_TIME: Final[int] = 3600  # 1 hour

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class ArtifactRevisionService:
    _artifact_repository: ArtifactRepository
    _artifact_registry_repository: ArtifactRegistryRepository
    _object_storage_repository: ObjectStorageRepository
    _vfs_storage_repository: VFSStorageRepository
    _storage_namespace_repository: StorageNamespaceRepository
    _huggingface_registry_repository: HuggingFaceRepository
    _reservoir_registry_repository: ReservoirRegistryRepository
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider
    _background_task_manager: BackgroundTaskManager

    def __init__(
        self,
        artifact_repository: ArtifactRepository,
        artifact_registry_repository: ArtifactRegistryRepository,
        object_storage_repository: ObjectStorageRepository,
        vfs_storage_repository: VFSStorageRepository,
        storage_namespace_repository: StorageNamespaceRepository,
        huggingface_registry_repository: HuggingFaceRepository,
        reservoir_registry_repository: ReservoirRegistryRepository,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        valkey_artifact_client: ValkeyArtifactDownloadTrackingClient,
        background_task_manager: BackgroundTaskManager,
    ) -> None:
        self._artifact_repository = artifact_repository
        self._artifact_registry_repository = artifact_registry_repository
        self._object_storage_repository = object_storage_repository
        self._vfs_storage_repository = vfs_storage_repository
        self._storage_namespace_repository = storage_namespace_repository
        self._huggingface_registry_repository = huggingface_registry_repository
        self._reservoir_registry_repository = reservoir_registry_repository
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._valkey_artifact_client = valkey_artifact_client
        self._background_task_manager = background_task_manager

    def _resolve_storage_namespace(self, reservoir_config: ReservoirConfig) -> str:
        """Resolve namespace based on storage type
        Args:
            reservoir_config: Reservoir configuration
        Returns:
            Namespace (bucket name for object storage, subpath for VFS storage)
        """
        match reservoir_config.config.storage_type:
            case ArtifactStorageType.OBJECT_STORAGE.value:
                return cast(ReservoirObjectStorageConfig, reservoir_config.config).bucket_name
            case ArtifactStorageType.VFS_STORAGE.value:
                return cast(ReservoirVFSStorageConfig, reservoir_config.config).subpath
            case _:
                raise UnsupportedStorageTypeError(
                    f"Unsupported storage type: {reservoir_config.config.storage_type}"
                )

    async def _get_storage_info(
        self, storage_name: str, namespace: str
    ) -> tuple[str, uuid.UUID, str]:
        """Get storage info by trying object_storage first, then vfs_storage as fallback
        Args:
            storage_name: Name of the storage
            namespace: Bucket name for object storage or subpath for VFS storage
        Returns: (storage_host, namespace_id, storage_name)
        """
        try:
            object_storage_data = await self._object_storage_repository.get_by_name(storage_name)
            storage_namespace = (
                await self._storage_namespace_repository.get_by_storage_and_namespace(
                    object_storage_data.id, namespace
                )
            )
            return object_storage_data.host, storage_namespace.id, object_storage_data.name
        except Exception:
            vfs_storage_data = await self._vfs_storage_repository.get_by_name(storage_name)
            storage_namespace = (
                await self._storage_namespace_repository.get_by_storage_and_namespace(
                    vfs_storage_data.id, namespace
                )
            )
            return vfs_storage_data.host, storage_namespace.id, vfs_storage_data.name

    async def get(self, action: GetArtifactRevisionAction) -> GetArtifactRevisionActionResult:
        revision = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )
        return GetArtifactRevisionActionResult(revision=revision)

    async def get_readme(
        self, action: GetArtifactRevisionReadmeAction
    ) -> GetArtifactRevisionReadmeActionResult:
        readme = await self._artifact_repository.get_artifact_revision_readme(
            action.artifact_revision_id
        )
        readme_data = ArtifactRevisionReadme(readme=readme)
        return GetArtifactRevisionReadmeActionResult(readme_data=readme_data)

    async def get_download_progress(
        self, action: GetDownloadProgressAction
    ) -> GetDownloadProgressActionResult:
        # 1. Get artifact_revision info
        revision = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )

        # 2. Get artifact info to extract model_id and revision
        artifact = await self._artifact_repository.get_artifact_by_id(revision.artifact_id)
        model_id = artifact.name  # artifact name is the model_id
        revision_name = revision.version

        # 3. Get local progress from valkey
        local_progress = await self._valkey_artifact_client.get_download_progress(
            model_id=model_id,
            revision=revision_name,
        )

        # Build local progress with status
        local_download_progress = ArtifactRevisionDownloadProgress(
            progress=local_progress if local_progress.artifact_progress else None,
            status=revision.status.value,
        )

        # 4. Query remote progress when delegation is enabled
        remote_download_progress: Optional[ArtifactRevisionDownloadProgress] = None

        # Only set remote for RESERVOIR type
        if artifact.registry_type == ArtifactRegistryType.RESERVOIR:
            reservoir_cfg = self._config_provider.config.reservoir
            # Default: create remote progress object with None progress and remote_status
            remote_status = revision.remote_status.value if revision.remote_status else "UNSCANNED"

            # Check if delegation is enabled
            if reservoir_cfg and reservoir_cfg.use_delegation:
                # If local is PULLING, return remote status without making remote request
                if revision.status == ArtifactStatus.PULLING:
                    remote_download_progress = ArtifactRevisionDownloadProgress(
                        progress=None,
                        status=remote_status,
                    )
                # Otherwise, query remote only if local has no progress
                elif local_progress.artifact_progress is None:
                    try:
                        # Get reservoir registry data
                        registry_data = await self._reservoir_registry_repository.get_reservoir_registry_data_by_id(
                            artifact.registry_id
                        )

                        # Create remote reservoir client
                        remote_reservoir_client = ReservoirRegistryClient(
                            registry_data=registry_data
                        )

                        # Query remote reservoir manager for download progress
                        remote_resp = await remote_reservoir_client.get_download_progress(
                            artifact_revision_id=action.artifact_revision_id,
                        )

                        # Parse response - expecting GetDownloadProgressResponse structure
                        # We want the remote's "local" progress as our "remote" progress
                        remote_local = remote_resp.download_progress.local

                        if remote_local:
                            remote_download_progress = ArtifactRevisionDownloadProgress(
                                progress=remote_local.progress,
                                status=remote_local.status,
                            )
                        else:
                            # Remote response exists but no local data
                            remote_download_progress = ArtifactRevisionDownloadProgress(
                                progress=None,
                                status=remote_status,
                            )
                    except Exception as e:
                        # If remote query fails, return remote status without progress
                        log.warning("Failed to get remote download progress {}", e)
                        remote_download_progress = ArtifactRevisionDownloadProgress(
                            progress=None,
                            status=remote_status,
                        )
                else:
                    # Local has progress, don't query remote
                    remote_download_progress = ArtifactRevisionDownloadProgress(
                        progress=None,
                        status=remote_status,
                    )
            else:
                # No delegation but still RESERVOIR type
                remote_download_progress = ArtifactRevisionDownloadProgress(
                    progress=None,
                    status=remote_status,
                )
        # else: Not RESERVOIR type, remote_download_progress remains None

        # 5. Build combined response
        combined_progress = CombinedDownloadProgress(
            local=local_download_progress,
            remote=remote_download_progress,
        )

        return GetDownloadProgressActionResult(download_progress=combined_progress)

    async def list_revision(
        self, action: ListArtifactRevisionsAction
    ) -> ListArtifactRevisionsActionResult:
        (
            artifacts_data,
            total_count,
        ) = await self._artifact_repository.list_artifact_revisions_paginated(
            pagination=action.pagination,
            ordering=action.ordering,
            filters=action.filters,
        )
        return ListArtifactRevisionsActionResult(data=artifacts_data, total_count=total_count)

    async def approve(
        self, action: ApproveArtifactRevisionAction
    ) -> ApproveArtifactRevisionActionResult:
        result = await self._artifact_repository.approve_artifact(action.artifact_revision_id)
        return ApproveArtifactRevisionActionResult(result=result)

    async def reject(
        self, action: RejectArtifactRevisionAction
    ) -> RejectArtifactRevisionActionResult:
        result = await self._artifact_repository.reject_artifact(action.artifact_revision_id)
        return RejectArtifactRevisionActionResult(result=result)

    async def associate_with_storage(
        self, action: AssociateWithStorageAction
    ) -> AssociateWithStorageActionResult:
        result = await self._artifact_repository.associate_artifact_with_storage(
            action.artifact_revision_id, action.storage_namespace_id, action.storage_type
        )
        return AssociateWithStorageActionResult(result=result)

    async def disassociate_with_storage(
        self, action: DisassociateWithStorageAction
    ) -> DisassociateWithStorageActionResult:
        result = await self._artifact_repository.disassociate_artifact_with_storage(
            action.artifact_revision_id, action.storage_namespace_id
        )
        return DisassociateWithStorageActionResult(result=result)

    async def cancel_import(self, action: CancelImportAction) -> CancelImportActionResult:
        await self._artifact_repository.reset_artifact_revision_status(action.artifact_revision_id)
        revision_data = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )
        return CancelImportActionResult(result=revision_data)

    async def import_revision(
        self, action: ImportArtifactRevisionAction
    ) -> ImportArtifactRevisionActionResult:
        try:
            revision_data = await self._artifact_repository.get_artifact_revision_by_id(
                action.artifact_revision_id
            )
            artifact = await self._artifact_repository.get_artifact_by_id(revision_data.artifact_id)

            reservoir_config = self._config_provider.config.reservoir
            if reservoir_config is None:
                raise ServerMisconfiguredError("Reservoir configuration is missing")

            storage_type = reservoir_config.config.storage_type
            reservoir_archive_storage = reservoir_config.archive_storage

            # Get bucket name or subpath depending on storage type
            namespace = self._resolve_storage_namespace(reservoir_config)
            storage_host, namespace_id, _ = await self._get_storage_info(
                reservoir_archive_storage, namespace
            )

            storage_proxy_client = self._storage_manager.get_manager_facing_client(storage_host)
            task_id: UUID
            match artifact.registry_type:
                case ArtifactRegistryType.HUGGINGFACE:
                    huggingface_registry_data = await self._huggingface_registry_repository.get_registry_data_by_artifact_id(
                        artifact.id
                    )

                    # Check current commit hash for this revision
                    commit_hash_resp = await storage_proxy_client.get_huggingface_model_commit_hash(
                        path=HuggingFaceGetCommitHashReqPathParam(
                            model_id=artifact.name,
                        ),
                        query=HuggingFaceGetCommitHashReqQueryParam(
                            revision=revision_data.version,
                            registry_name=huggingface_registry_data.name,
                        ),
                    )
                    latest_commit_hash = commit_hash_resp.commit_hash

                    # Skip import if artifact revision is already Available and commit hash matches
                    # If current_commit_hash is None, always proceed with import
                    if self._is_latest_commit_hash(revision_data, latest_commit_hash):
                        # Return early without calling import API
                        return ImportArtifactRevisionActionResult(
                            result=revision_data, task_id=None
                        )

                    await self._artifact_repository.update_artifact_revision_status(
                        action.artifact_revision_id, ArtifactStatus.PULLING
                    )

                    huggingface_result = await storage_proxy_client.import_huggingface_models(
                        HuggingFaceImportModelsReq(
                            models=[
                                ModelTarget(model_id=artifact.name, revision=revision_data.version)
                            ],
                            registry_name=huggingface_registry_data.name,
                            storage_step_mappings=reservoir_config.resolve_storage_step_selection(),
                        )
                    )
                    task_id = huggingface_result.task_id
                case ArtifactRegistryType.RESERVOIR:
                    registry_data = (
                        await self._reservoir_registry_repository.get_registry_data_by_artifact_id(
                            artifact.id
                        )
                    )

                    async def _task(reporter: ProgressReporter) -> None:
                        # When use_delegation is True, if remote status is not AVAILABLE, delegate import
                        if reservoir_config.use_delegation:
                            remote_reservoir_client = ReservoirRegistryClient(
                                registry_data=registry_data
                            )

                            remote_progress_status_resp = (
                                await remote_reservoir_client.get_download_progress(
                                    revision_data.id
                                )
                            )
                            remote_progress = remote_progress_status_resp.download_progress.local

                            # Wait for remote to start pulling if not yet AVAILABLE
                            if remote_progress.status != ArtifactStatus.AVAILABLE.value:
                                delegate_import_req = DelegateImportArtifactsReq(
                                    artifact_revision_ids=[revision_data.id],
                                    artifact_type=artifact.type,
                                    # We can't pass delegatee_reservoir_id here since we don't have it.
                                    # So we depend on default.
                                    delegator_reservoir_id=None,
                                    delegatee_target=None,
                                )

                                await remote_reservoir_client.delegate_import_artifacts(
                                    delegate_import_req
                                )

                                # Poll until remote status is AVAILABLE
                                elapsed_time = 0

                                log.info(
                                    "Waiting for remote artifact to become AVAILABLE. "
                                    "artifact_revision_id: {}",
                                    revision_data.id,
                                )

                                while elapsed_time < _REMOTE_ARTIFACT_MAX_WAIT_TIME:
                                    await asyncio.sleep(_REMOTE_ARTIFACT_STATUS_POLL_INTERVAL)
                                    elapsed_time += _REMOTE_ARTIFACT_STATUS_POLL_INTERVAL

                                    remote_progress_status_resp = (
                                        await remote_reservoir_client.get_download_progress(
                                            revision_data.id
                                        )
                                    )
                                    remote_progress = (
                                        remote_progress_status_resp.download_progress.local
                                    )

                                    if remote_progress.status == ArtifactStatus.AVAILABLE.value:
                                        log.info(
                                            "Remote artifact is now AVAILABLE. "
                                            "artifact_revision_id: {}, elapsed_time: {}s",
                                            revision_data.id,
                                            elapsed_time,
                                        )
                                        break

                                    log.info(
                                        "Waiting for remote artifact. Status: {}, Elapsed: {}s, "
                                        "artifact_revision_id: {}",
                                        remote_progress.status,
                                        elapsed_time,
                                        revision_data.id,
                                    )
                                else:
                                    # Timeout reached
                                    raise RemoteReservoirArtifactImportError(
                                        f"Timeout waiting for remote artifact to become AVAILABLE. "
                                        f"artifact_revision_id: {revision_data.id}, "
                                        f"Last status: {remote_progress.status}"
                                    )

                        await self._artifact_repository.update_artifact_revision_status(
                            action.artifact_revision_id, ArtifactStatus.PULLING
                        )

                        # TODO: Utilize this internal import task_id
                        _result = await storage_proxy_client.import_reservoir_models(
                            ReservoirImportModelsReq(
                                models=[
                                    ModelTarget(
                                        model_id=artifact.name, revision=revision_data.version
                                    )
                                ],
                                registry_name=registry_data.name,
                                storage_step_mappings=reservoir_config.resolve_storage_step_selection(),
                            )
                        )

                    task_id = await self._background_task_manager.start(_task)
                case _:
                    raise InvalidArtifactRegistryTypeError(
                        f"Unsupported artifact registry type: {artifact.registry_type}"
                    )

            await self.associate_with_storage(
                AssociateWithStorageAction(
                    revision_data.id, namespace_id, ArtifactStorageType(storage_type)
                )
            )

        except Exception as e:
            await self._artifact_repository.update_artifact_revision_status(
                action.artifact_revision_id, ArtifactStatus.FAILED
            )
            raise e

        return ImportArtifactRevisionActionResult(result=revision_data, task_id=task_id)

    async def cleanup(
        self, action: CleanupArtifactRevisionAction
    ) -> CleanupArtifactRevisionActionResult:
        revision_data = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )

        if revision_data.status in [ArtifactStatus.SCANNED, ArtifactStatus.PULLING]:
            raise ArtifactDeletionBadRequestError(
                "Artifact revision status not ready to be deleted"
            )

        await self._artifact_repository.reset_artifact_revision_status(revision_data.id)
        artifact_data = await self._artifact_repository.get_artifact_by_id(
            revision_data.artifact_id
        )

        reservoir_config = self._config_provider.config.reservoir
        if reservoir_config is None:
            raise ServerMisconfiguredError("Reservoir configuration is missing")

        reservoir_archive_storage = reservoir_config.archive_storage
        # TODO: Abstract this.
        namespace = self._resolve_storage_namespace(reservoir_config)

        storage_data = None
        vfs_storage_data = None
        storage_namespace = None
        storage_host = None
        namespace_id = None
        storage_name = None

        try:
            storage_data = await self._object_storage_repository.get_by_name(
                reservoir_archive_storage
            )
            storage_namespace = (
                await self._storage_namespace_repository.get_by_storage_and_namespace(
                    storage_data.id, namespace
                )
            )
            storage_host = storage_data.host
            namespace_id = storage_namespace.id
            storage_name = storage_data.name
        except Exception:
            vfs_storage_data = await self._vfs_storage_repository.get_by_name(
                reservoir_archive_storage
            )
            storage_namespace = (
                await self._storage_namespace_repository.get_by_storage_and_namespace(
                    vfs_storage_data.id, namespace
                )
            )
            storage_host = vfs_storage_data.host
            namespace_id = storage_namespace.id
            storage_name = vfs_storage_data.name

        if storage_data is None and vfs_storage_data is None:
            raise ArtifactDeletionError("Storage data not found for artifact deletion")

        storage_proxy_client = self._storage_manager.get_manager_facing_client(storage_host)
        key = f"{artifact_data.name}/{revision_data.version}/"

        try:
            await storage_proxy_client.delete_s3_object(
                storage_name=storage_name,
                bucket_name=storage_namespace.namespace,
                req=DeleteObjectReq(
                    key=key,
                ),
            )
        except Exception as e:
            raise ArtifactDeletionError("Failed to delete artifact from storage") from e

        await self.disassociate_with_storage(
            DisassociateWithStorageAction(
                artifact_revision_id=revision_data.id,
                storage_namespace_id=namespace_id,
            )
        )

        artifact_revision = await self._artifact_repository.get_artifact_revision_by_id(
            revision_data.id
        )
        return CleanupArtifactRevisionActionResult(result=artifact_revision)

    async def delegate_import_revision_batch(
        self, action: DelegateImportArtifactRevisionBatchAction
    ) -> DelegateImportArtifactRevisionBatchActionResult:
        # If this is a leaf node, perform local import instead of delegation
        reservoir_cfg = self._config_provider.config.reservoir
        if reservoir_cfg is None:
            raise ServerMisconfiguredError("Reservoir configuration is missing")

        if not reservoir_cfg.use_delegation:
            registry_id = None
            if action.delegatee_target:
                registry_id = action.delegatee_target.target_registry_id

            registry_meta = await self._resolve_artifact_registry_meta(
                action.artifact_type, registry_id
            )

            try:
                # TODO: Improve this
                task_ids = []
                result: list[ArtifactRevisionData] = []
                for revision_id in action.artifact_revision_ids:
                    import_result = await self.import_revision(
                        ImportArtifactRevisionAction(artifact_revision_id=revision_id)
                    )
                    task_ids.append(import_result.task_id)  # Keep None values for zip alignment
                    result.append(import_result.result)
            except Exception as e:
                raise RemoteReservoirArtifactImportError(
                    f"Failed to import artifacts from remote reservoir: {e}"
                ) from e

            return DelegateImportArtifactRevisionBatchActionResult(result=result, task_ids=task_ids)

        # If not a leaf node, perform delegation to remote reservoir
        registry_meta = await self._resolve_artifact_registry_meta(
            action.artifact_type, action.delegator_reservoir_id
        )
        registry_type = registry_meta.type
        registry_id = registry_meta.registry_id

        if registry_type != ArtifactRegistryType.RESERVOIR:
            raise ArtifactImportBadRequestError(
                "Only Reservoir type registry is supported for delegated import"
            )

        registry_data = await self._reservoir_registry_repository.get_reservoir_registry_data_by_id(
            registry_id
        )

        # Update remote_status to SCANNED for all revisions before delegation
        result_revisions: list[ArtifactRevisionData] = []
        for revision_id in action.artifact_revision_ids:
            revision_data = await self._artifact_repository.get_artifact_revision_by_id(revision_id)
            result_revisions.append(revision_data)

        # Pass delegatee_reservoir_id to delegator_reservoir_id for the remote call
        delegatee_reservoir_id = (
            action.delegatee_target.delegatee_reservoir_id if action.delegatee_target else None
        )
        req = DelegateImportArtifactsReq(
            artifact_revision_ids=action.artifact_revision_ids,
            delegator_reservoir_id=delegatee_reservoir_id,
            delegatee_target=action.delegatee_target,
            artifact_type=action.artifact_type,
        )

        remote_reservoir_client = ReservoirRegistryClient(registry_data=registry_data)
        client_resp = await remote_reservoir_client.delegate_import_artifacts(req)

        # Extract task_ids from remote response
        task_ids = [uuid.UUID(task.task_id) if task.task_id else None for task in client_resp.tasks]

        return DelegateImportArtifactRevisionBatchActionResult(
            result=result_revisions, task_ids=task_ids
        )

    async def _resolve_artifact_registry_meta(
        self, artifact_type: Optional[ArtifactType], registry_id_or_none: Optional[uuid.UUID]
    ) -> ArtifactRegistryData:
        if registry_id_or_none is None:
            artifact_registry_cfg = self._config_provider.config.artifact_registry
            if artifact_registry_cfg is None:
                raise ServerMisconfiguredError("Artifact registry configuration is missing.")

            # TODO: Handle `artifact_type` for other types
            registry_name = artifact_registry_cfg.model_registry
            registry_meta = (
                await self._artifact_registry_repository.get_artifact_registry_data_by_name(
                    registry_name
                )
            )
        else:
            registry_id = registry_id_or_none
            registry_meta = await self._artifact_registry_repository.get_artifact_registry_data(
                registry_id
            )

        return registry_meta

    def _is_latest_commit_hash(
        self, artifact_revision: ArtifactRevisionData, latest_commit_hash: Optional[str]
    ) -> bool:
        """
        Used in huggingface import to check if the latest commit hash matches the stored digest.
        """
        if latest_commit_hash is None:
            return False
        if (
            artifact_revision.status == ArtifactStatus.AVAILABLE
            and artifact_revision.digest == latest_commit_hash
        ):
            return True
        return False



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Optional, Self

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from ai.backend.manager.data.permission.permission import PermissionCreator, PermissionData
from ai.backend.manager.data.permission.types import (
    EntityType,
    OperationType,
)

from ...base import (
    GUID,
    Base,
    IDColumn,
    StrEnumType,
)

if TYPE_CHECKING:
    from .permission_group import PermissionGroupRow


class PermissionRow(Base):
    __tablename__ = "permissions"
    __table_args__ = (sa.Index("ix_id_permission_group_id", "id", "permission_group_id"),)

    id: uuid.UUID = IDColumn()
    permission_group_id: uuid.UUID = sa.Column("permission_group_id", GUID, nullable=False)
    entity_type: EntityType = sa.Column(
        "entity_type", StrEnumType(EntityType, length=32), nullable=False
    )
    operation: OperationType = sa.Column(
        "operation", StrEnumType(OperationType, length=32), nullable=False
    )

    permission_group_row: Optional[PermissionGroupRow] = relationship(
        "PermissionGroupRow",
        back_populates="permission_rows",
        primaryjoin="PermissionGroupRow.id == foreign(PermissionRow.permission_group_id)",
    )

    @classmethod
    def from_input(cls, input: PermissionCreator) -> Self:
        return cls(
            permission_group_id=input.permission_group_id,
            entity_type=input.entity_type,
            operation=input.operation,
        )

    def to_data(self) -> PermissionData:
        return PermissionData(
            id=self.id,
            permission_group_id=self.permission_group_id,
            entity_type=self.entity_type,
            operation=self.operation,
        )



"""GraphQL scaling group module."""

from .resolver import all_scaling_groups_v2, scaling_groups_v2
from .types import (
    ScalingGroupFilterGQL,
    ScalingGroupOrderByGQL,
    ScalingGroupOrderFieldGQL,
    ScalingGroupV2GQL,
)

__all__ = (
    # Types
    "ScalingGroupV2GQL",
    "ScalingGroupFilterGQL",
    "ScalingGroupOrderByGQL",
    "ScalingGroupOrderFieldGQL",
    # Resolvers
    "scaling_groups_v2",
    "all_scaling_groups_v2",
)



from typing import Optional

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError, KeypairResourcePolicyNotFound
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.resource.types import KeyPairResourcePolicyData
from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.keypair_resource_policy.actions.modify_keypair_resource_policy import (
    KeyPairResourcePolicyModifier,
)
from ai.backend.manager.services.keypair_resource_policy.types import KeyPairResourcePolicyCreator

keypair_resource_policy_db_source_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.DB_SOURCE, layer=LayerType.KEYPAIR_RESOURCE_POLICY_DB_SOURCE
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=5,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class KeypairResourcePolicyDBSource:
    """Database source for KeyPair resource policy operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @keypair_resource_policy_db_source_resilience.apply()
    async def insert(self, creator: KeyPairResourcePolicyCreator) -> KeyPairResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            db_row = KeyPairResourcePolicyRow.from_creator(creator)
            db_sess.add(db_row)
            await db_sess.flush()
            # Refresh the object to ensure all attributes are loaded
            await db_sess.refresh(db_row)
            return db_row.to_dataclass()

    @keypair_resource_policy_db_source_resilience.apply()
    async def update(
        self, name: str, modifier: KeyPairResourcePolicyModifier
    ) -> KeyPairResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            check_query = sa.select(KeyPairResourcePolicyRow).where(
                KeyPairResourcePolicyRow.name == name
            )
            existing_row = await db_sess.scalar(check_query)
            if existing_row is None:
                raise KeypairResourcePolicyNotFound(
                    f"Keypair resource policy with name {name} not found."
                )
            fields = modifier.fields_to_update()
            update_stmt = (
                sa.update(KeyPairResourcePolicyRow)
                .where(KeyPairResourcePolicyRow.name == name)
                .values(**fields)
                .returning(KeyPairResourcePolicyRow)
            )
            query_stmt = (
                sa.select(KeyPairResourcePolicyRow)
                .from_statement(update_stmt)
                .execution_options(populate_existing=True)
            )
            updated_row: Optional[KeyPairResourcePolicyRow] = await db_sess.scalar(query_stmt)
            if updated_row is None:
                raise KeypairResourcePolicyNotFound(
                    f"Keypair resource policy with name {name} not found after update."
                )
            return updated_row.to_dataclass()

    @keypair_resource_policy_db_source_resilience.apply()
    async def delete(self, name: str) -> KeyPairResourcePolicyData:
        """Deletes a keypair resource policy by name."""
        async with self._db.begin_session() as db_sess:
            delete_stmt = (
                sa.delete(KeyPairResourcePolicyRow)
                .where(KeyPairResourcePolicyRow.name == name)
                .returning(KeyPairResourcePolicyRow)
            )
            query_stms = (
                sa.select(KeyPairResourcePolicyRow)
                .from_statement(delete_stmt)
                .execution_options(populate_existing=True)
            )
            row: Optional[KeyPairResourcePolicyRow] = await db_sess.scalar(query_stms)
            if row is None:
                raise KeypairResourcePolicyNotFound(
                    f"Keypair resource policy with name {name} not found."
                )
            await db_sess.delete(row)
            return row.to_dataclass()



from __future__ import annotations

from datetime import datetime, tzinfo
from decimal import Decimal
from enum import StrEnum
from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence
from uuid import UUID

import attrs
import msgpack
import sqlalchemy as sa
from sqlalchemy.orm import joinedload, load_only

from ai.backend.common.types import SlotName
from ai.backend.common.utils import nmget
from ai.backend.manager.data.kernel.types import KernelStatus

if TYPE_CHECKING:
    from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient

from .group import GroupRow
from .kernel import LIVE_STATUS, RESOURCE_USAGE_KERNEL_STATUSES, KernelRow
from .session import SessionRow
from .user import UserRow
from .utils import ExtendedAsyncSAEngine

__all__: Sequence[str] = (
    "ResourceGroupUnit",
    "ResourceUsage",
    "BaseResourceUsageGroup",
    "parse_resource_usage",
    "parse_resource_usage_groups",
    "fetch_resource_usage",
)


class ResourceGroupUnit(StrEnum):
    KERNEL = "kernel"
    SESSION = "session"
    PROJECT = "project"
    DOMAIN = "domain"
    TOTAL = "total"


@attrs.define(slots=True)
class ResourceUsage:
    nfs: set = attrs.field(factory=set)
    cpu_allocated: float = attrs.field(default=0.0)
    cpu_used: float = attrs.field(default=0.0)
    mem_allocated: int = attrs.field(default=0)
    mem_used: int = attrs.field(default=0)
    shared_memory: int = attrs.field(default=0)
    disk_allocated: int = attrs.field(default=0)  # TODO: disk quota limit
    disk_used: int = attrs.field(default=0)
    io_read: int = attrs.field(default=0)
    io_write: int = attrs.field(default=0)
    device_type: set = attrs.field(factory=set)
    smp: float = attrs.field(default=0.0)
    gpu_mem_allocated: float = attrs.field(default=0.0)
    gpu_allocated: float = attrs.field(default=0.0)

    agent_ids: set[str] = attrs.field(factory=set)

    def __add__(self, other: Any) -> ResourceUsage:
        if not isinstance(other, ResourceUsage):
            raise TypeError(
                "ResourceUsage should only be added to `ResourceUsage` type,"
                f" not `{type(other)}` type."
            )
        self.nfs |= other.nfs  # Is this correct??
        self.device_type |= other.device_type
        self.cpu_allocated += other.cpu_allocated
        self.cpu_used += other.cpu_used
        self.mem_allocated += other.mem_allocated
        self.mem_used += other.mem_used
        self.shared_memory += other.shared_memory
        self.disk_allocated += other.disk_allocated
        self.disk_used += other.disk_used
        self.io_read += other.io_read
        self.io_write += other.io_write
        self.smp += other.smp
        self.gpu_mem_allocated += other.gpu_mem_allocated
        self.gpu_allocated += other.gpu_allocated
        self.agent_ids |= other.agent_ids
        return self

    def to_json(self) -> Mapping[str, Any]:
        return {
            "nfs": list(self.nfs),
            "cpu_allocated": self.cpu_allocated,
            "cpu_used": self.cpu_used,
            "mem_allocated": self.mem_allocated,
            "mem_used": self.mem_used,
            "shared_memory": self.shared_memory,
            "disk_allocated": self.disk_allocated,
            "disk_used": self.disk_used,
            "io_read": self.io_read,
            "io_write": self.io_write,
            "device_type": list(self.device_type),
            "smp": self.smp,
            "gpu_mem_allocated": self.gpu_mem_allocated,
            "gpu_allocated": self.gpu_allocated,
        }

    def copy(self) -> ResourceUsage:
        return attrs.evolve(self, nfs={*self.nfs}, device_type={*self.device_type})


def to_str(val: Any) -> Optional[str]:
    return str(val) if val is not None else None


@attrs.define(slots=True)
class BaseResourceUsageGroup:
    group_unit: ResourceGroupUnit
    child_usage_group: dict = attrs.field(factory=dict)

    project_row: Optional[GroupRow] = attrs.field(default=None)
    session_row: Optional[SessionRow] = attrs.field(default=None)
    kernel_row: Optional[KernelRow] = attrs.field(default=None)

    created_at: Optional[datetime] = attrs.field(default=None)
    terminated_at: Optional[datetime] = attrs.field(default=None)
    scheduled_at: Optional[str] = attrs.field(default=None)
    used_time: Optional[str] = attrs.field(default=None)
    used_days: Optional[int] = attrs.field(default=None)
    agents: Optional[set[str]] = attrs.field(default=None)

    user_id: Optional[UUID] = attrs.field(default=None)
    user_email: Optional[str] = attrs.field(default=None)
    full_name: Optional[str] = attrs.field(default=None)  # User's full_name
    access_key: Optional[UUID] = attrs.field(default=None)
    project_id: Optional[UUID] = attrs.field(default=None)
    project_name: Optional[str] = attrs.field(default=None)
    kernel_id: Optional[UUID] = attrs.field(default=None)
    container_ids: Optional[set[str]] = attrs.field(default=None)
    session_id: Optional[UUID] = attrs.field(default=None)
    session_name: Optional[str] = attrs.field(default=None)
    domain_name: Optional[str] = attrs.field(default=None)
    images: Optional[set[str]] = attrs.field(default=None)

    last_stat: Optional[Mapping[str, Any]] = attrs.field(default=None)
    extra_info: Mapping[str, Any] = attrs.field(factory=dict)

    status: Optional[str] = attrs.field(default=None)
    status_info: Optional[str] = attrs.field(default=None)
    status_history: Optional[str] = attrs.field(default=None)
    cluster_mode: Optional[str] = attrs.field(default=None)

    total_usage: ResourceUsage = attrs.field(factory=ResourceUsage)

    # def to_json(self, child: bool = False) -> dict[str, Any]:
    #     return_val = {
    #         **self.to_map(),
    #         "agents": self.total_usage.agent_ids,
    #         "group_unit": self.group_unit.value,
    #         "total_usage": self.total_usage.to_json(),
    #     }
    #     if child:
    #         belonged_infos: dict[str, list[Mapping[str, Any]]] = {}
    #         for g in self.child_usage_group.values():
    #             if g.group_unit.value not in belonged_infos:
    #                 belonged_infos[g.group_unit.value] = []
    #             belonged_infos[g.group_unit.value].append(g.to_json())
    #         return_val = {
    #             **belonged_infos,
    #             **return_val,
    #             # **{k:v for k, v in belonged_infos.items()},
    #         }
    #     return return_val

    def to_map(self) -> dict[str, Any]:
        return {
            "created_at": self.created_at,
            "terminated_at": self.terminated_at,
            "used_time": self.used_time,
            "used_days": self.used_days,
            "user_id": self.user_id,
            "user_email": self.user_email,
            "access_key": self.access_key,
            "project_id": self.project_id,
            "project_name": self.project_name,
            "kernel_id": self.kernel_id,
            "container_ids": self.container_ids.copy() if self.container_ids is not None else None,
            "session_id": self.session_id,
            "session_name": self.session_name,
            "domain_name": self.domain_name,
            "last_stat": {**self.last_stat} if self.last_stat is not None else None,
            "extra_info": {**self.extra_info} if self.extra_info is not None else None,
            "full_name": self.full_name,
            "images": self.images.copy() if self.images is not None else None,
            "agents": self.agents.copy() if self.agents is not None else None,
            "status": self.status,
            "status_info": self.status_info,
            "status_history": self.status_history,
            "cluster_mode": self.cluster_mode,
            "scheduled_at": self.scheduled_at,
            "total_usage": self.total_usage.copy(),
        }

    def to_json_base(self) -> dict[str, Any]:
        return {
            "created_at": to_str(self.created_at),
            "used_time": self.used_time,
            "used_days": self.used_days,
            "user_id": to_str(self.user_id),
            "user_email": self.user_email,
            "access_key": self.access_key,
            "project_id": to_str(self.project_id),
            "project_name": self.project_name,
            "kernel_id": to_str(self.kernel_id),
            "container_ids": list(self.container_ids) if self.container_ids is not None else [],
            "session_id": to_str(self.session_id),
            "session_name": self.session_name,
            "domain_name": self.domain_name,
            "last_stat": self.last_stat,
            "extra_info": self.extra_info,
            "full_name": to_str(self.full_name),
            "images": list(self.images) if self.images is not None else [],
            "agents": list(self.agents) if self.agents is not None else [],
            "status": to_str(self.status),
            "status_info": to_str(self.status_info),
            "status_history": to_str(self.status_history),
            "cluster_mode": to_str(self.cluster_mode),
            "scheduled_at": to_str(self.scheduled_at),
            "terminated_at": to_str(self.terminated_at),
            "total_usage": self.total_usage.to_json(),
        }


@attrs.define(slots=True, kw_only=True)
class KernelResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.KERNEL
    # child_usage_group: dict = attrs.field(factory=dict)
    agent: str
    kernel_id: UUID
    project_row: GroupRow
    session_row: SessionRow
    kernel_row: KernelRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return {
            **self.to_json_base(),
            "agents": list(self.agents) if self.agents is not None else [],
            "agent": self.agent,
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> KernelResourceUsage:
        if usage_group.group_unit != ResourceGroupUnit.KERNEL:
            raise ValueError(
                "Unable to parse `KernelResourceUsage` from usage_group "
                "that DOES NOT have `ResourceGroupUnit.KERNEL` on group_unit field."
            )
        if (
            usage_group.project_row is None
            or usage_group.session_row is None
            or usage_group.kernel_row is None
        ):
            raise ValueError(
                "Cannot parse `KernelResourceUsage` from usage_group that have None value field"
            )
        return cls(
            project_row=usage_group.project_row,
            session_row=usage_group.session_row,
            kernel_row=usage_group.kernel_row,
            agent=usage_group.kernel_row.agent,
            **usage_group.to_map(),
        )

    def register_resource_group(
        self,
        other: BaseResourceUsageGroup,
    ) -> bool:
        return True


@attrs.define(slots=True, kw_only=True)
class SessionResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.SESSION
    child_usage_group: dict[UUID, KernelResourceUsage] = attrs.Factory(dict)
    session_id: UUID
    project_row: GroupRow
    session_row: SessionRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return_val = {
            **self.to_json_base(),
            "agents": list(self.agents) if self.agents is not None else [],
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }
        if child:
            return_val = {
                **return_val,
                "kernels": {
                    str(g.kernel_id): g.to_json(child) for g in self.child_usage_group.values()
                },
            }
        return return_val

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> SessionResourceUsage:
        if usage_group.group_unit not in (ResourceGroupUnit.KERNEL, ResourceGroupUnit.SESSION):
            raise ValueError(
                "Unable to parse `SessionResourceUsage` from usage_group that DOES NOT have"
                " `ResourceGroupUnit.KERNEL` or `ResourceGroupUnit.SESSION` on group_unit field."
            )
        if usage_group.project_row is None or usage_group.session_row is None:
            raise ValueError(
                "Cannot parse `SessionResourceUsage` from usage_group that have None value field"
            )
        return cls(
            project_row=usage_group.project_row,
            session_row=usage_group.session_row,
            **{
                **usage_group.to_map(),
                "total_usage": ResourceUsage(),
            },
        )

    def register_resource_group(self, other: BaseResourceUsageGroup) -> bool:
        if other.kernel_id is None:
            return False
        if other.kernel_id in self.child_usage_group:
            return False
        self.child_usage_group[other.kernel_id] = KernelResourceUsage.from_base_usage_group(other)
        is_registered = self.child_usage_group[other.kernel_id].register_resource_group(other)
        if is_registered:
            self.total_usage += other.total_usage
            if self.images is None:
                self.images = {*(other.images or set())}
            else:
                self.images |= other.images or set()
            if self.agents is None:
                self.agents = {*(other.agents or set())}
            else:
                self.agents |= other.agents or set()
            if self.container_ids is None:
                self.container_ids = {*(other.container_ids or set())}
            else:
                self.container_ids |= other.container_ids or set()
        return is_registered


@attrs.define(slots=True, kw_only=True)
class ProjectResourceUsage(BaseResourceUsageGroup):
    group_unit: ResourceGroupUnit = ResourceGroupUnit.PROJECT
    child_usage_group: dict[UUID, SessionResourceUsage] = attrs.Factory(dict)
    project_id: UUID
    project_row: GroupRow

    def to_json(self, child: bool = False) -> dict[str, Any]:
        return_val = {
            **self.to_json_base(),
            "agents": list(self.total_usage.agent_ids),
            "group_unit": self.group_unit.value,
            "total_usage": self.total_usage.to_json(),
        }
        if child:
            return_val = {
                **return_val,
                "sessions": {
                    str(g.session_id): g.to_json(child) for g in self.child_usage_group.values()
                },
            }
        return return_val

    @classmethod
    def from_base_usage_group(cls, usage_group: BaseResourceUsageGroup) -> ProjectResourceUsage:
        if usage_group.project_row is None:
            raise ValueError(
                "Cannot parse `ProjectResourceUsage` from usage_group that have None value field"
            )
        filtered_data = {
            **usage_group.to_map(),
            "images": None,
            "full_name": None,
            "status": None,
            "status_info": None,
            "status_history": None,
            "cluster_mode": None,
            "scheduled_at": None,
            "terminated_at": None,
            "total_usage": ResourceUsage(),
        }
        return cls(
            project_row=usage_group.project_row,
            **filtered_data,
        )

    def register_resource_group(self, other: BaseResourceUsageGroup) -> bool:
        if other.session_id is None:
            return False
        if other.session_id not in self.child_usage_group:
            self.child_usage_group[other.session_id] = SessionResourceUsage.from_base_usage_group(
                other
            )
        is_registered = self.child_usage_group[other.session_id].register_resource_group(other)
        if is_registered:
            self.total_usage += other.total_usage
            if self.images is None:
                self.images = {*(other.images or set())}
            else:
                self.images |= other.images or set()
            if self.agents is None:
                self.agents = {*(other.agents or set())}
            else:
                self.agents |= other.agents or set()
            if self.container_ids is None:
                self.container_ids = {*(other.container_ids or set())}
            else:
                self.container_ids |= other.container_ids or set()
        return is_registered


# @attrs.define(slots=True, kw_only=True)
# class DomainResourceUsage(BaseResourceUsageGroup):
#     group_unit: ResourceGroupUnit = attrs.field(default=ResourceGroupUnit.DOMAIN)
#     child_usage_group: dict[UUID, ProjectResourceUsage] = attrs.Factory(dict)


def parse_total_resource_group(
    groups: Sequence[BaseResourceUsageGroup],
) -> tuple[dict[UUID, ProjectResourceUsage], ResourceUsage]:
    group_map: dict[UUID, ProjectResourceUsage] = {}
    total_usage = ResourceUsage()
    for g in groups:
        key = g.project_id
        if key is None:
            continue
        if key not in group_map:
            group_map[key] = ProjectResourceUsage.from_base_usage_group(g)
        is_registered = group_map[key].register_resource_group(g)
        if is_registered:
            total_usage += g.total_usage
    return group_map, total_usage


def parse_resource_usage(
    kernel: KernelRow,
    last_stat: Optional[Mapping[str, Any]],
) -> ResourceUsage:
    if not last_stat:
        return ResourceUsage(
            agent_ids={kernel.agent},
        )
    nfs = set()
    if kernel.vfolder_mounts:
        # For >=22.03, return used host directories instead of volume host, which is not so useful.
        nfs = set([str(mount.host_path) for mount in kernel.vfolder_mounts])
    elif kernel.mounts and isinstance(kernel.mounts[0], list):
        # For the kernel records that have legacy contents of `mounts`.
        nfs = set([mount[2] for mount in kernel.mounts])

    device_type = set()
    smp = 0
    gpu_mem_allocated = 0
    if kernel.attached_devices and "cuda" in kernel.attached_devices:
        for dev_info in kernel.attached_devices["cuda"]:
            if (model_name := dev_info.get("model_name")) is not None:
                device_type.add(model_name)
            smp += int(nmget(dev_info, "data.smp", 0))
            gpu_mem_allocated += int(nmget(dev_info, "data.mem", 0))
    gpu_allocated = Decimal(0)
    for key, value in kernel.occupied_slots.items():
        if SlotName(key).is_accelerator():
            gpu_allocated += value

    return ResourceUsage(
        agent_ids={kernel.agent},
        nfs={*nfs},
        cpu_allocated=float(kernel.occupied_slots.get("cpu", 0)),
        cpu_used=float(nmget(last_stat, "cpu_used.current", 0)),
        mem_allocated=int(kernel.occupied_slots.get("mem", 0)),
        mem_used=int(nmget(last_stat, "mem.capacity", 0)),
        shared_memory=int(nmget(kernel.resource_opts, "shmem", 0)),
        disk_allocated=0,
        disk_used=int(nmget(last_stat, "io_scratch_size/stats.max", 0, "/")),
        io_read=int(nmget(last_stat, "io_read.current", 0)),
        io_write=int(nmget(last_stat, "io_write.current", 0)),
        device_type={*device_type},
        smp=float(smp),
        gpu_mem_allocated=float(gpu_mem_allocated),
        gpu_allocated=float(gpu_allocated),
    )


async def parse_resource_usage_groups(
    kernels: list[KernelRow],
    valkey_stat_client: ValkeyStatClient,
    local_tz: tzinfo,
) -> list[BaseResourceUsageGroup]:
    stat_map = {k.id: k.last_stat for k in kernels}
    stat_empty_kerns = [k.id for k in kernels if not k.last_stat]

    kernel_ids_str = [str(kern_id) for kern_id in stat_empty_kerns]
    raw_stats = await valkey_stat_client.get_user_kernel_statistics_batch(kernel_ids_str)
    for kern_id, raw_stat in zip(stat_empty_kerns, raw_stats):
        if raw_stat is None:
            continue
        stat_map[kern_id] = msgpack.unpackb(raw_stat)

    return [
        BaseResourceUsageGroup(
            kernel_row=kern,
            project_row=kern.session.group,
            session_row=kern.session,
            created_at=kern.created_at,
            terminated_at=kern.terminated_at,
            scheduled_at=kern.status_history.get(KernelStatus.SCHEDULED.name),
            used_time=kern.used_time,
            used_days=kern.get_used_days(local_tz),
            last_stat=stat_map[kern.id],
            user_id=kern.session.user_uuid,
            user_email=kern.session.user.email if kern.session.user is not None else None,
            access_key=kern.session.access_key,
            project_id=kern.session.group.id if kern.session.group is not None else None,
            project_name=kern.session.group.name if kern.session.group is not None else None,
            kernel_id=kern.id,
            container_ids={kern.container_id},
            session_id=kern.session_id,
            session_name=kern.session.name,
            domain_name=kern.session.domain_name,
            full_name=kern.session.user.full_name if kern.session.user is not None else None,
            images={kern.image},
            agents={kern.agent},
            status=kern.status.name,
            status_history=kern.status_history,
            cluster_mode=kern.cluster_mode,
            status_info=kern.status_info,
            group_unit=ResourceGroupUnit.KERNEL,
            total_usage=parse_resource_usage(kern, stat_map[kern.id]),
        )
        for kern in kernels
    ]


SESSION_RESOURCE_SELECT_COLS = (
    SessionRow.created_at,
    SessionRow.user_uuid,
    SessionRow.name,
    SessionRow.domain_name,
    SessionRow.id,
    SessionRow.group_id,
    SessionRow.access_key,
    SessionRow.images,
    SessionRow.cluster_mode,
    SessionRow.status_history,
    SessionRow.status,
    SessionRow.status_info,
)

PROJECT_RESOURCE_SELECT_COLS = (
    GroupRow.created_at,
    GroupRow.id,
    GroupRow.name,
    GroupRow.domain_name,
)

KERNEL_RESOURCE_SELECT_COLS = (
    KernelRow.agent,
    KernelRow.created_at,
    KernelRow.terminated_at,
    KernelRow.last_stat,
    KernelRow.status_history,
    KernelRow.status,
    KernelRow.status_info,
    KernelRow.session_id,
    KernelRow.id,
    KernelRow.container_id,
    KernelRow.occupied_slots,
    KernelRow.attached_devices,
    KernelRow.vfolder_mounts,
    KernelRow.mounts,
    KernelRow.resource_opts,
    KernelRow.image,
    KernelRow.cluster_mode,
)


def _parse_query(
    kernel_cond: Optional[sa.sql.BinaryExpression] = None,
    session_cond: Optional[sa.sql.BinaryExpression] = None,
    project_cond: Optional[sa.sql.BinaryExpression] = None,
) -> sa.sql.Select:
    session_load = joinedload(KernelRow.session)
    if session_cond is not None:
        session_load = joinedload(KernelRow.session.and_(session_cond))

    project_load = joinedload(SessionRow.group)
    if project_cond is not None:
        project_load = joinedload(SessionRow.group.and_(project_cond))
    query = sa.select(KernelRow).options(
        load_only(*KERNEL_RESOURCE_SELECT_COLS),
        session_load.options(
            load_only(*SESSION_RESOURCE_SELECT_COLS),
            joinedload(SessionRow.user).options(
                load_only(UserRow.email, UserRow.username, UserRow.full_name)
            ),
            project_load.options(load_only(*PROJECT_RESOURCE_SELECT_COLS)),
        ),
    )
    if kernel_cond is not None:
        query = query.where(kernel_cond)
    return query


async def fetch_resource_usage(
    db_engine: ExtendedAsyncSAEngine,
    start_date: datetime,
    end_date: datetime,
    session_ids: Optional[Sequence[UUID]] = None,
    project_ids: Optional[Sequence[UUID]] = None,
) -> list[KernelRow]:
    project_cond = None
    if project_ids:
        project_cond = GroupRow.id.in_(project_ids)
    session_cond = None
    if session_ids:
        session_cond = SessionRow.id.in_(session_ids)
    query = _parse_query(
        kernel_cond=(
            # Filter sessions which existence period overlaps with requested period
            (
                (KernelRow.terminated_at >= start_date)
                & (KernelRow.created_at < end_date)
                & (KernelRow.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))
            )
            |
            # Or, filter running sessions which created before requested end_date
            ((KernelRow.created_at < end_date) & (KernelRow.status.in_(LIVE_STATUS)))
        ),
        session_cond=session_cond,
        project_cond=project_cond,
    )
    async with db_engine.begin_readonly_session() as db_sess:
        result = await db_sess.execute(query)
        kernels = result.scalars().all()

    return kernels



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional

from ai.backend.common.types import DefaultForUnspecified, ResourceSlot


@dataclass
class UserResourcePolicyData:
    name: str
    max_vfolder_count: int
    max_quota_scope_size: int
    max_session_count_per_model_session: int
    max_customized_image_count: int


@dataclass
class ProjectResourcePolicyData:
    name: str
    max_vfolder_count: int
    max_quota_scope_size: int
    max_network_count: int


@dataclass
class KeyPairResourcePolicyData:
    name: str
    created_at: datetime = field(compare=False)
    default_for_unspecified: DefaultForUnspecified
    total_resource_slots: ResourceSlot
    max_session_lifetime: int
    max_concurrent_sessions: int
    max_pending_session_count: Optional[int]
    max_pending_session_resource_slots: Optional[Any]  # TODO: Use ResourceSlot.
    max_concurrent_sftp_sessions: int
    max_containers_per_session: int
    idle_timeout: int
    allowed_vfolder_hosts: dict[str, Any]


@dataclass
class ScalingGroupProxyTarget:
    addr: str
    api_token: str



from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    Optional,
    Sequence,
    cast,
    override,
)

import aiohttp
import graphene
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import joinedload, load_only, selectinload

from ai.backend.common.types import (
    HardwareMetadata,
    VFolderHostPermission,
    VFolderHostPermissionMap,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.storage_proxy.session_manager import (
    AUTH_TOKEN_HDR,
    StorageProxyInfo,
    StorageSessionManager,
    VolumeInfo,
)  # For compatibility with existing code

from .base import Item, PaginatedList
from .rbac import (
    AbstractPermissionContext,
    AbstractPermissionContextBuilder,
    DomainScope,
    ProjectScope,
    ScopeType,
    UserScope,
    get_predefined_roles_in_scope,
)
from .rbac.context import ClientContext
from .rbac.permission_defs import StorageHostPermission
from .resource_policy import KeyPairResourcePolicyRow
from .user import UserRow

if TYPE_CHECKING:
    from .gql import GraphQueryContext

# Left this for compatibility with existing code
__all__ = (
    "StorageProxyInfo",
    "VolumeInfo",
    "StorageSessionManager",
    "StorageVolume",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class StorageVolume(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    # id: {proxy_name}:{name}
    backend = graphene.String()
    fsprefix = graphene.String()
    path = graphene.String()
    capabilities = graphene.List(graphene.String)
    hardware_metadata = graphene.JSONString()
    performance_metric = graphene.JSONString()
    usage = graphene.JSONString()
    proxy = graphene.String(
        description="Added in 24.03.0. Name of the proxy which this volume belongs to."
    )
    name = graphene.String(description="Added in 24.03.0. Name of the storage.")

    async def resolve_hardware_metadata(self, info: graphene.ResolveInfo) -> HardwareMetadata:
        ctx: GraphQueryContext = info.context
        return await ctx.registry.gather_storage_hwinfo(self.id)

    async def resolve_performance_metric(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        try:
            proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(self.id)
            manager_client = ctx.storage_manager.get_manager_facing_client(proxy_name)
            storage_reply = await manager_client.get_volume_performance_metric(volume_name)
            return storage_reply["metric"]
        except aiohttp.ClientResponseError:
            return {}

    async def resolve_usage(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(self.id)
        try:
            manager_client = ctx.storage_manager.get_manager_facing_client(proxy_name)
            return await manager_client.get_fs_usage(volume_name)
        except aiohttp.ClientResponseError:
            return {}

    @classmethod
    def from_info(cls, proxy_name: str, volume_info: VolumeInfo) -> StorageVolume:
        return cls(
            id=f"{proxy_name}:{volume_info['name']}",
            backend=volume_info["backend"],
            path=volume_info["path"],
            fsprefix=volume_info["fsprefix"],
            capabilities=volume_info["capabilities"],
            name=volume_info["name"],
            proxy=proxy_name,
        )

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        filter: Optional[str] = None,
    ) -> int:
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return len(volumes)

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[StorageVolume]:
        # For consistency we add filter/order params here, but it's actually noop.
        if filter is not None or order is not None:
            log.warning(
                "Paginated list of storage volumes igonores custom filtering and/or ordering"
            )
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return [
            cls.from_info(proxy_name, volume_info)
            for proxy_name, volume_info in volumes[offset : offset + limit]
        ]

    @classmethod
    async def load_by_id(
        cls,
        ctx: GraphQueryContext,
        id: str,
    ) -> StorageVolume:
        proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(id)
        try:
            proxy_info = ctx.storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        async with proxy_info.session.request(
            "GET",
            proxy_info.manager_api_url / "volumes",
            raise_for_status=True,
            headers={AUTH_TOKEN_HDR: proxy_info.secret},
        ) as resp:
            reply = await resp.json()
            for volume_data in reply["volumes"]:
                if volume_data["name"] == volume_name:
                    return cls.from_info(proxy_name, volume_data)
            else:
                raise ValueError(
                    f"no such volume in the storage proxy {proxy_name!r}: {volume_name!r}",
                )


class StorageVolumeList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(StorageVolume, required=True)


# RBAC

ALL_STORAGE_HOST_PERMISSIONS: frozenset[StorageHostPermission] = frozenset([
    perm for perm in StorageHostPermission
])
OWNER_PERMISSIONS: frozenset[StorageHostPermission] = ALL_STORAGE_HOST_PERMISSIONS
ADMIN_PERMISSIONS: frozenset[StorageHostPermission] = ALL_STORAGE_HOST_PERMISSIONS
MONITOR_PERMISSIONS: frozenset[StorageHostPermission] = ALL_STORAGE_HOST_PERMISSIONS
PRIVILEGED_MEMBER_PERMISSIONS: frozenset[StorageHostPermission] = ALL_STORAGE_HOST_PERMISSIONS
MEMBER_PERMISSIONS: frozenset[StorageHostPermission] = ALL_STORAGE_HOST_PERMISSIONS

LEGACY_VFHOST_PERMISSION_TO_HOST__PERMISSION_MAP: Mapping[
    VFolderHostPermission, frozenset[StorageHostPermission]
] = {
    VFolderHostPermission.CREATE: frozenset([StorageHostPermission.CREATE_FOLDER]),
    VFolderHostPermission.MODIFY: frozenset([StorageHostPermission.UPDATE_ATTRIBUTE]),
    VFolderHostPermission.DELETE: frozenset([
        StorageHostPermission.DELETE_VFOLDER,
        StorageHostPermission.DELETE_CONTENT,
    ]),
    VFolderHostPermission.MOUNT_IN_SESSION: frozenset([
        StorageHostPermission.MOUNT_RO,
        StorageHostPermission.MOUNT_RW,
        StorageHostPermission.MOUNT_WD,
    ]),
    VFolderHostPermission.UPLOAD_FILE: frozenset([
        StorageHostPermission.READ_ATTRIBUTE,
        StorageHostPermission.WRITE_CONTENT,
        StorageHostPermission.DELETE_CONTENT,
    ]),
    VFolderHostPermission.DOWNLOAD_FILE: frozenset([
        StorageHostPermission.READ_ATTRIBUTE,
        StorageHostPermission.WRITE_CONTENT,
        StorageHostPermission.DELETE_CONTENT,
    ]),
    VFolderHostPermission.INVITE_OTHERS: frozenset([
        StorageHostPermission.ASSIGN_PERMISSION_TO_OTHERS
    ]),
    VFolderHostPermission.SET_USER_PERM: frozenset([
        StorageHostPermission.ASSIGN_PERMISSION_TO_OTHERS
    ]),
}

ALL_LEGACY_VFHOST_PERMISSIONS = {perm for perm in VFolderHostPermission}


def _legacy_vf_perms_to_host_rbac_perms(
    perms: list[VFolderHostPermission],
) -> frozenset[StorageHostPermission]:
    if set(perms) == ALL_LEGACY_VFHOST_PERMISSIONS:
        return ALL_STORAGE_HOST_PERMISSIONS
    result: frozenset[StorageHostPermission] = frozenset()
    for perm in perms:
        result |= LEGACY_VFHOST_PERMISSION_TO_HOST__PERMISSION_MAP[perm]
    return result


StorageHostToPermissionMap = Mapping[str, frozenset[StorageHostPermission]]


@dataclass
class PermissionContext(AbstractPermissionContext[StorageHostPermission, str, str]):
    @property
    def host_to_permissions_map(self) -> StorageHostToPermissionMap:
        return self.object_id_to_additional_permission_map

    async def build_query(self) -> sa.sql.Select | None:
        return None

    async def calculate_final_permission(self, rbac_obj: str) -> frozenset[StorageHostPermission]:
        host_name = rbac_obj
        return self.object_id_to_additional_permission_map.get(host_name, frozenset())


class PermissionContextBuilder(
    AbstractPermissionContextBuilder[StorageHostPermission, PermissionContext]
):
    db_session: SASession

    def __init__(self, db_session: SASession) -> None:
        self.db_session = db_session

    @override
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[StorageHostPermission]:
        roles = await get_predefined_roles_in_scope(ctx, target_scope, self.db_session)
        permissions = await self._calculate_permission_by_predefined_roles(roles)
        return permissions

    @override
    async def build_ctx_in_system_scope(
        self,
        ctx: ClientContext,
    ) -> PermissionContext:
        from .domain import DomainRow

        perm_ctx = PermissionContext()
        _domain_query_stmt = sa.select(DomainRow).options(load_only(DomainRow.name))
        for row in await self.db_session.scalars(_domain_query_stmt):
            to_be_merged = await self.build_ctx_in_domain_scope(ctx, DomainScope(row.name))
            perm_ctx.merge(to_be_merged)
        return perm_ctx

    @override
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> PermissionContext:
        from .domain import DomainRow

        permissions = await self.calculate_permission(ctx, scope)
        if not permissions:
            # User is not part of the domain.
            return PermissionContext()

        stmt = (
            sa.select(DomainRow)
            .where(DomainRow.name == scope.domain_name)
            .options(load_only(DomainRow.allowed_vfolder_hosts))
        )
        domain_row = cast(DomainRow | None, await self.db_session.scalar(stmt))
        if domain_row is None:
            return PermissionContext()
        host_permissions = cast(VFolderHostPermissionMap, domain_row.allowed_vfolder_hosts)
        result = PermissionContext(
            object_id_to_additional_permission_map={
                host: _legacy_vf_perms_to_host_rbac_perms(perms)
                for host, perms in host_permissions.items()
            }
        )
        return result

    @override
    async def build_ctx_in_project_scope(
        self,
        ctx: ClientContext,
        scope: ProjectScope,
    ) -> PermissionContext:
        from .group import GroupRow

        permissions = await self.calculate_permission(ctx, scope)
        if not permissions:
            # User is not part of the domain.
            return PermissionContext()

        stmt = (
            sa.select(GroupRow)
            .where(GroupRow.id == scope.project_id)
            .options(load_only(GroupRow.allowed_vfolder_hosts))
        )
        project_row = cast(GroupRow | None, await self.db_session.scalar(stmt))
        if project_row is None:
            return PermissionContext()
        host_permissions = cast(VFolderHostPermissionMap, project_row.allowed_vfolder_hosts)
        result = PermissionContext(
            object_id_to_additional_permission_map={
                host: _legacy_vf_perms_to_host_rbac_perms(perms)
                for host, perms in host_permissions.items()
            }
        )
        return result

    @override
    async def build_ctx_in_user_scope(
        self,
        ctx: ClientContext,
        scope: UserScope,
    ) -> PermissionContext:
        from .keypair import KeyPairRow

        permissions = await self.calculate_permission(ctx, scope)
        if not permissions:
            # User is not part of the domain.
            return PermissionContext()
        stmt = (
            sa.select(UserRow)
            .where(UserRow.uuid == scope.user_id)
            .options(
                selectinload(UserRow.keypairs).options(
                    joinedload(KeyPairRow.resource_policy).options(
                        load_only(KeyPairResourcePolicyRow.allowed_vfolder_hosts)
                    )
                )
            )
        )
        user_row = cast(UserRow | None, await self.db_session.scalar(stmt))
        if user_row is None:
            return PermissionContext()

        object_id_to_additional_permission_map: defaultdict[
            str, frozenset[StorageHostPermission]
        ] = defaultdict(frozenset)

        for keypair in user_row.keypairs:
            resource_policy = cast(KeyPairResourcePolicyRow | None, keypair.resource_policy)
            if resource_policy is None:
                continue
            host_permissions = cast(VFolderHostPermissionMap, resource_policy.allowed_vfolder_hosts)
            for host, perms in host_permissions.items():
                object_id_to_additional_permission_map[host] |= _legacy_vf_perms_to_host_rbac_perms(
                    perms
                )

        result = PermissionContext(
            object_id_to_additional_permission_map=object_id_to_additional_permission_map
        )
        return result

    @override
    @classmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[StorageHostPermission]:
        return OWNER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[StorageHostPermission]:
        return ADMIN_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[StorageHostPermission]:
        return MONITOR_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[StorageHostPermission]:
        return PRIVILEGED_MEMBER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[StorageHostPermission]:
        return MEMBER_PERMISSIONS



"""Add digest to artifact_revisions

Revision ID: 5a5142f6d251
Revises: d811b103dbfc
Create Date: 2025-10-31 03:54:59.846269

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5a5142f6d251"
down_revision = "d811b103dbfc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("artifact_revisions", sa.Column("digest", sa.String(), nullable=True))


def downgrade() -> None:
    op.drop_column("artifact_revisions", "digest")




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/manager:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI Manager',
    'entry_points': {
        'backendai_cli_v10': [
            'mgr = ai.backend.manager.cli.__main__:main',
            'mgr.start-server = ai.backend.manager.server:main',
        ],
        'backendai_scheduler_v10': [
            'fifo = ai.backend.manager.scheduler.fifo:FIFOSlotScheduler',
            'lifo = ai.backend.manager.scheduler.fifo:LIFOSlotScheduler',
            'drf = ai.backend.manager.scheduler.drf:DRFScheduler',
        ],
        'backendai_agentselector_v10': [
            'legacy = ai.backend.manager.scheduler.agent_selector:LegacyAgentSelector',
            'roundrobin = ai.backend.manager.scheduler.agent_selector:RoundRobinAgentSelector',
            'concentrated = ai.backend.manager.scheduler.agent_selector:ConcentratedAgentSelector',
            'dispersed = ai.backend.manager.scheduler.agent_selector:DispersedAgentSelector',
        ],
        'backendai_error_monitor_v20': [
            'intrinsic = ai.backend.manager.plugin.error_monitor:ErrorMonitor',
        ],
        'backendai_event_dispatcher_v20': [
            'intrinsic = ai.backend.manager.plugin.error_monitor:ErrorEventDispatcher',
        ],
        'backendai_network_manager_v1': [
            'overlay = ai.backend.manager.network.overlay:OverlayNetworkPlugin',
        ],
    },
    'install_requires': (
        'Jinja2~=3.1.6',
        'PyJWT~=2.10.1',
        'PyYAML~=6.0',
        'SQLAlchemy[postgresql_asyncpg]~=1.4.54',
        'aiodataloader~=0.4.2',
        'aiodocker==0.24.0',
        'aiofiles~=24.1.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp_sse>=2.2',
        'aiohttp~=3.12.15',
        'aiomonitor~=0.7.0',
        'aiotools~=2.2.3',
        'alembic~=1.13.2',
        'async_timeout~=4.0',
        'attrs>=25.3',
        """backend.ai-cli==25.18.0rc2
""",
        """backend.ai-common==25.18.0rc2
""",
        """backend.ai-logging==25.18.0rc2
""",
        """backend.ai-plugin==25.18.0rc2
""",
        'bcrypt~=4.2.0',
        'boto3~=1.35',
        'callosum~=1.0.3',
        'click~=8.1.7',
        'cryptography>=44.0.2',
        'etcd-client-py~=0.4.1',
        'graphene-federation~=3.2.0',
        'graphene~=3.3.0',
        'graphql-core~=3.2.6',
        'graphql-relay~=3.2.0',
        'lark~=1.1.5',
        'more-itertools~=10.5.0',
        'msgpack~=1.1.0',
        'multidict~=6.6.4',
        'prometheus-client~=0.21.1',
        'pydantic[email]~=2.11.3',
        'python-dateutil>=2.9',
        'pyzmq~=26.4',
        'redis[hiredis]==4.5.5',
        'ruamel.yaml~=0.18.10',
        'setproctitle~=1.3.5',
        'strawberry-graphql~=0.278.0',
        'tabulate~=0.8.9',
        'tenacity>=9.0',
        'tomli~=2.0.1',
        'trafaret~=2.1',
        'typeguard~=4.3',
        'types-Jinja2',
        'types-PyYAML',
        'types-aiofiles',
        'types-python-dateutil',
        'types-redis',
        'types-six',
        'types-tabulate',
        'typing_extensions~=4.11',
        'uvloop~=0.22.1; sys_platform != "Windows"',
        'valkey-glide~=2.0.1',
        'yarl~=1.19.0',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI Manager

← [Back to Architecture Overview](../README.md#manager)

## Purpose

The Manager is the central orchestrator of the Backend.AI cluster. It schedules computing sessions (sessions and kernels), allocates resources, manages the lifecycle of sessions, and provides API gateway functionality through REST and GraphQL interfaces.

## Key Responsibilities

### 1. API Gateway
- Provide REST API and GraphQL API to clients
- Request authentication and authorization
- Rate limiting and quota management
- API versioning management

### 2. Session Scheduling
- Allocate computing resources for user compute session requests
- Select optimal agents based on various scheduling algorithms
- Manage session lifecycle (creation, execution, termination)
- Cluster-mode session scheduling and management

### 3. Resource Management
- Track cluster-wide resource status
- Collect and aggregate agent resource information
- Manage resource allocation and release
- Handle resource presets and quotas

### 4. User and Organization Management
- User account and authentication management
- Group and domain organization management
- Permission and role-based access control
- Credential (access key/secret key) management

### 5. Virtual Folder Management
- Provide persistent storage to users
- Integrate various storage backends
- Manage file upload/download
- Manage folder sharing and permissions

### 6. Image Registry Management
- Manage container image repositories
- Scan and synchronize image metadata
- Validate and manage image versions
- Control allowed images per domain

## Entry Points

The Manager accepts and processes external requests through 4 entry points.

### 1. REST API

**Framework**: aiohttp (async HTTP web framework)

**Location**: `src/ai/backend/manager/api/`

**Key Features**:
- HTTP/HTTPS-based communication
- JSON request/response format
- JWT or API Key-based authentication
- Validation and authentication via middleware stack

**Processing Flow**:
```
HTTP Request → Middleware Stack → REST Handler → Action Processor → Service → Repository → DB
```

**Related Documentation**: [REST API Documentation](./api/README.md)

### 2. GraphQL API

**Framework**: Strawberry (current) + Graphene (Legacy, DEPRECATED)

**Location**:
- Strawberry: `src/ai/backend/manager/api/gql/`
- Graphene (Legacy): `src/ai/backend/manager/models/gql_models/`

**Related Documentation**:
- [GraphQL API (Strawberry)](./api/gql/README.md)
- [Legacy GraphQL (Graphene)](./models/gql_models/README.md) - DEPRECATED

### 3. Event Dispatcher

**Framework**: Backend.AI Event Dispatcher

**Location**: `src/ai/backend/common/events/`

**Event Types**:
- **Broadcast Events**: Received by all Manager instances
- **Anycast Events**: Received by only one Manager instance

**Processing Flow**:
```
Event Producer → Event Dispatcher → Event Handler → Service
```

**Related Documentation**: [Event Dispatcher System](../common/events/README.md)

### 4. Background Task Handler

**Framework**: Backend.AI Background Task Handler

**Location**: `src/ai/backend/common/bgtask/`

**Purpose**:
Handles long-running tasks asynchronously. Issues Task IDs that allow clients to subscribe to progress updates or results via events.

**Processing Flow**:
```
Task Request → Background Task Handler → Task Execute → Event Notification
```

**Related Documentation**: [Background Task Handler System](../common/bgtask/README.md)

### Entry Point Interactions

Each entry point operates independently, but service logic can trigger background tasks or publish events as needed.

**Interaction Examples**:
```
REST API Handler → Service Logic → Event Publish (notify other Manager instances)
Event Handler → Service Logic → Background Task Trigger (when async processing needed)
```

**Integrated Architecture**:

```
┌─────────────┐  ┌─────────────┐  ┌─────────────┐  ┌─────────────┐
│  REST API   │  │ GraphQL API │  │   Event     │  │ Background  │
│  (aiohttp)  │  │(Strawberry) │  │ Dispatcher  │  │    Task     │
└──────┬──────┘  └──────┬──────┘  └──────┬──────┘  └──────┬──────┘
       │                │                │                │
       │                │                │                │
       └────────────────┴────────────────┴────────────────┘
                              │
                    ┌─────────▼──────────┐
                    │  Services Layer    │
                    └─────────┬──────────┘
                              │
                    ┌─────────▼──────────┐
                    │ Repositories Layer │
                    └────────────────────┘
```

## Architecture

```
┌───────────────────────────────────────────┐
│              API Layer                    │
│  - REST API Handler (aiohttp)             │
│  - GraphQL Handler (strawberry)           │
│  - Authentication & Authorization         │
│  - Request Validation                     │
└──────────────────┬────────────────────────┘
                   │
┌──────────────────┴────────────────────────┐
│            Actions Layer                  │
│  - Session Lifecycle Actions              │
│  - Resource Allocation Actions            │
│  - User Management Actions                │
│  - VFolder Management Actions             │
└──────────────────┬────────────────────────┘
                   │
┌──────────────────┴────────────────────────┐
│           Services Layer                  │
│  - Scheduling Service                     │
│  - Session Management Service             │
│  - Resource Quota Service                 │
│  - Event Processing Service               │
└──────────────────┬────────────────────────┘
                   │
┌──────────────────┴────────────────────────┐
│         Repositories Layer                │
│  - Session Repository                     │
│  - Agent Repository                       │
│  - User Repository                        │
│  - VFolder Repository                     │
└──────────────────┬────────────────────────┘
                   │
┌──────────────────┴────────────────────────┐
│            Models Layer                   │
│  - SQLAlchemy ORM Models                  │
│  - Domain Types                           │
└───────────────────────────────────────────┘
```

## Directory Structure

```
manager/
├── models/              # Database schema and ORM models
│   ├── alembic/        # Database migration scripts
│   ├── user.py         # User and credential models
│   ├── session.py      # Session and kernel models
│   ├── agent.py        # Agent models
│   ├── vfolder.py      # Virtual folder models
│   ├── scaling_group.py # Scaling group models
│   └── ...
├── repositories/        # Data access layer
│   ├── session/        # Session data access
│   ├── agent/          # Agent data access
│   ├── user/           # User data access
│   └── ...
├── services/            # Business logic layer
│   ├── session/        # Session lifecycle management
│   └── ...
├── api/                 # API handlers and routes
│   ├── gql/             # GraphQL schema and resolvers
│   ├── auth.py          # Authentication handlers
│   └── ...
├── config/              # Configuration management
├── cli/                 # CLI commands
│   ├── fixture.py       # Test data management
│   └── ...
├── clients/             # External service clients
│   ├── agent/           # Agent RPC client
│   ├── storage_proxy/   # Storage proxy client
│   └── ...
├── scheduler/           # Scheduling algorithms and logic
│   ├── dispatcher.py   # Scheduling dispatcher
│   ├── predicates.py   # Scheduling predicates
│   └── ...
├── server.py            # Main server entry point
└── defs.py              # Shared constants and types
```

## Core Concepts

### Sessions
Sessions represent user compute requests and are composed of one or more kernels:
- **Session ID**: Unique identifier for the session
- **Access Key**: Owner's API access key
- **Status**: Session state (PENDING, RUNNING, TERMINATED, etc.)
- **Resource Allocation**: Allocated CPU, memory, GPU resources
- **Cluster Size**: Number of kernels in multi-container mode

Session Lifecycle:
1. User creates session via API
2. Manager schedules and allocates resources
3. Agent creates and starts kernel containers
4. User executes code and uses services
5. Session terminates (user request or timeout)
6. Resources are returned to the pool

### Agents
Agents are compute node workers that execute kernels:
- **Agent ID**: Unique identifier for the agent
- **Status**: Agent state (ALIVE, LOST, TERMINATED)
- **Available Resources**: CPU, memory, GPU capacity
- **Occupied Slots**: Currently allocated resources
- **Scaling Group**: Group to which the agent belongs

Agent Monitoring:
- Manager periodically checks agent health
- Agent heartbeat and resource status updates
- Automatic detection of lost agents
- Kernel rebalancing during failures

### Scaling Groups
Scaling groups manage agent clusters logically:
- **Group Name**: Unique identifier for the scaling group
- **Scheduler Type**: Scheduling algorithm (FIFO, LIFO, DRF, etc.)
- **Agent Members**: Agents belonging to the group
- **Allowed vFolder Hosts**: Permitted storage backends
- **Resource Limits**: Per-group resource limits

### Virtual Folders (VFolders)
VFolders provide persistent storage:
- **Folder Name**: User-defined folder name
- **Host**: Storage backend location
- **Ownership**: User or group ownership
- **Permissions**: Read/write permissions
- **Quota**: Storage capacity limit

## Infrastructure Dependencies

### Required Infrastructure

#### PostgreSQL (Persistent Data)
- **Purpose**:
  - Store all Backend.AI metadata
  - User/Group/Domain information
  - Session and kernel history
  - VFolder metadata
  - Resource allocation records
- **Halfstack Port**: 8101 (host) → 5432 (container)
- **Key Tables**:
  - `users`, `keypairs` - User credentials
  - `groups`, `domains` - Organization structure
  - `kernels`, `sessions` - Session information
  - `agents` - Agent status
  - `vfolders` - VFolder metadata
  - `scaling_groups` - Scaling group configuration

#### Redis (Caching and Real-time Data)
- **Purpose**:
  - Cache frequently accessed data
  - Distributed locking
  - Agent live status tracking
  - Session rate limiting
  - Temporary session data storage
- **Halfstack Port**: 8111 (host) → 6379 (container)

#### etcd (Global Configuration)
- **Purpose**:
  - Store cluster-wide configuration
  - Service discovery
  - Agent registration
  - Dynamic configuration updates
- **Halfstack Port**: 8121 (host) → 2379 (container)

### Optional Infrastructure (Observability)

#### Prometheus (Metrics Collection)
- **Purpose**:
  - API request metrics
  - Session scheduling metrics
  - Resource usage metrics
  - Background task metrics
- **Internal Port**: 18080 (separate from main API port 8091)
- **Exposed Endpoints**:
  - `http://localhost:18080/metrics` - Prometheus metrics endpoint
  - `http://localhost:18080/metrics/service_discovery` - Service discovery endpoint for automated metrics collection configuration
- **Key Metrics**:
  - `backendai_api_request_count` - Total API requests
  - `backendai_api_request_duration_sec` - Request processing time
  - `backendai_scheduler_enqueue_success` - Successful scheduling count
  - `backendai_agent_registry_count` - Number of agents
- **Note**: The service discovery endpoint provides automated configuration for Prometheus to discover all Backend.AI component endpoints

#### Loki (Log Aggregation)
- **Purpose**:
  - Session lifecycle events
  - API request/response logs
  - Scheduling decision logs
  - Error and exception logs
- **Log Labels**:
  - `component` - Component identifier (manager)
  - `session_id` - Session identifier
  - `user_id` - User identifier
  - `level` - Log level (info, warning, error)

#### Grafana (Visualization)
- **Purpose**:
  - Real-time metrics dashboards
  - Resource usage visualization
  - Session status monitoring
  - Alert management

## Configuration

See `configs/manager/halfstack.conf` for configuration file examples.

### Key Configuration Items

**Database Settings**:
- PostgreSQL connection string
- Connection pool size
- Query timeout settings

**Redis Settings**:
- Redis connection information
- Connection pool configuration

**etcd Settings**:
- etcd endpoint addresses
- Configuration key prefix (namespace)

**API Settings**:
- Listen address and port
- CORS configuration

**Scheduling Settings**:
- Default scheduler type
- Scheduling interval
- Resource allocation policy

### Halfstack Configuration

**Recommended**: Use the `./scripts/install-dev.sh` script for development environment setup.

#### Starting Development Environment
```bash
# Setup development environment via script (recommended)
./scripts/install-dev.sh

# Start Manager
./backend.ai mgr start-server
```

#### API Access
- REST API: http://localhost:8081
- GraphQL API: http://localhost:8081/graphql
- Admin GraphQL UI: http://localhost:8081/graphql-ui

## Metrics and Monitoring

### Prometheus Metrics

The Manager component exposes Prometheus metrics at the `/metrics` endpoint for monitoring system health and performance.

#### Label Conventions

Many metrics share common labels for error tracking and classification:

**Error-related Labels** (populated only when errors occur):
- `domain`: Error domain categorizing the error source (e.g., "session", "agent", "storage")
- `operation`: Specific operation that failed (e.g., "create", "terminate", "allocate")
- `error_detail`: Detailed error information for debugging

**Status Labels**:
- `status`: Operation outcome - typically "success" or "failure"
- `success`: Boolean string ("True" or "False") indicating operation success

#### API Metrics

REST API request monitoring metrics.

**`backendai_api_request_count`** (Counter)
- **Description**: Total number of REST API requests received
- **Labels**:
  - `method`: HTTP method (GET, POST, PUT, DELETE, PATCH)
  - `endpoint`: API endpoint path (e.g., "/v1/session/create")
  - `domain`: Error domain (empty if successful)
  - `operation`: Error operation (empty if successful)
  - `error_detail`: Error details (empty if successful)
  - `status_code`: HTTP response status code (200, 400, 500, etc.)

**`backendai_api_request_duration_sec`** (Histogram)
- **Description**: API request processing time in seconds
- **Labels**: Same as `backendai_api_request_count`

#### GraphQL Metrics

GraphQL query execution monitoring metrics.

**`backendai_graphql_request_count`** (Counter)
- **Description**: Total number of GraphQL queries executed
- **Labels**:
  - `operation_type`: GraphQL operation type (query, mutation, subscription)
  - `field_name`: GraphQL field being accessed
  - `parent_type`: Parent type in the GraphQL schema
  - `operation_name`: Named operation from the query
  - `domain`: Error domain (empty if successful)
  - `operation`: Error operation (empty if successful)
  - `error_detail`: Error details (empty if successful)
  - `success`: "True" or "False" indicating query success

**`backendai_graphql_request_duration_sec`** (Histogram)
- **Description**: GraphQL query processing time in seconds
- **Labels**: Same as `backendai_graphql_request_count`

#### Event Metrics

Internal event processing metrics.

**`backendai_event_count`** (Counter)
- **Description**: Total number of events processed
- **Labels**:
  - `event_type`: Type of event (e.g., "session_terminated", "kernel_started")

**`backendai_event_failure_count`** (Counter)
- **Description**: Number of failed event processing attempts
- **Labels**:
  - `event_type`: Type of event that failed
  - `exception`: Exception class name (e.g., "SessionNotFound", "AgentError")
  - `domain`: Error domain
  - `operation`: Error operation
  - `error_detail`: Error details

**`backendai_event_processing_time_sec`** (Histogram)
- **Description**: Event processing time in seconds
- **Labels**:
  - `event_type`: Type of event
  - `status`: "success" or "failure"
  - `domain`: Error domain (empty if successful)
  - `operation`: Error operation (empty if successful)
  - `error_detail`: Error details (empty if successful)

#### Background Task Metrics

Periodic and scheduled task execution metrics.

**`backendai_bgtask_count`** (Gauge)
- **Description**: Number of currently running background tasks
- **Labels**:
  - `task_name`: Background task identifier (e.g., "recalc_agent_resource_occupancy")

**`backendai_bgtask_done_count`** (Counter)
- **Description**: Total number of completed background tasks
- **Labels**:
  - `task_name`: Background task identifier
  - `status`: "success" or "failure"
  - `domain`: Error domain (empty if successful)
  - `operation`: Error operation (empty if successful)
  - `error_detail`: Error details (empty if successful)

**`backendai_bgtask_processing_time_sec`** (Histogram)
- **Description**: Background task execution time in seconds
- **Labels**: Same as `backendai_bgtask_done_count`

#### Action Metrics

High-level business operation metrics.

**`backendai_action_count`** (Counter)
- **Description**: Total number of actions executed
- **Labels**:
  - `entity_type`: Type of entity being operated on (e.g., "session", "kernel", "agent")
  - `operation_type`: Type of operation (e.g., "create", "terminate", "restart")
  - `status`: "success" or "failure"
  - `domain`: Error domain (empty if successful)
  - `operation`: Error operation (empty if successful)
  - `error_detail`: Error details (empty if successful)
- **Example**: Tracks session creation, termination, and other major lifecycle operations

**`backendai_action_duration_sec`** (Histogram)
- **Description**: Action execution time in seconds
- **Labels**: Same as `backendai_action_count`

#### Layer Operation Metrics

Granular metrics for operations at each architectural layer.

**`backendai_layer_operation_triggered_count`** (Gauge)
- **Description**: Number of layer operations currently in progress
- **Labels**:
  - `domain`: Domain type (valkey, repository, client)
  - `layer`: Specific layer (e.g., "session_repository", "agent_client", "valkey_live")
  - `operation`: Operation name (e.g., "fetch_session", "create_kernel")

**`backendai_layer_operation_count`** (Counter)
- **Description**: Total number of layer operations completed
- **Labels**:
  - `domain`: Domain type (valkey, repository, client)
  - `layer`: Specific layer identifier
  - `operation`: Operation name
  - `success`: "True" or "False"

**`backendai_layer_operation_error_count`** (Counter)
- **Description**: Total number of layer operation errors
- **Labels**:
  - `domain`: Domain type
  - `layer`: Specific layer identifier
  - `operation`: Operation name
  - `error_code`: Error code or "internal_error"

**`backendai_layer_retry_count`** (Counter)
- **Description**: Number of retries for layer operations
- **Labels**:
  - `domain`: Domain type
  - `layer`: Specific layer identifier
  - `operation`: Operation name

**`backendai_layer_operation_duration_sec`** (Histogram)
- **Description**: Layer operation execution time in seconds
- **Labels**:
  - `domain`: Domain type
  - `layer`: Specific layer identifier
  - `operation`: Operation name
  - `success`: "True" or "False"

#### System Metrics

System resource usage metrics.

**`backendai_async_task_count`** (Gauge)
- **Description**: Number of active asyncio tasks
- **Labels**: None

**`backendai_cpu_usage_percent`** (Gauge)
- **Description**: CPU usage percentage of the Manager process
- **Labels**: None

**`backendai_memory_used_rss`** (Gauge)
- **Description**: Resident Set Size (RSS) memory usage in bytes
- **Labels**: None

**`backendai_memory_used_vms`** (Gauge)
- **Description**: Virtual Memory Size (VMS) in bytes
- **Labels**: None

#### Sweeper Metrics

Resource cleanup and garbage collection metrics.

**`backendai_sweep_session_count`** (Counter)
- **Description**: Total number of session cleanup operations
- **Labels**:
  - `status`: Session status being cleaned (e.g., "TERMINATED", "ERROR")
  - `success`: "True" or "False"

**`backendai_sweep_kernel_count`** (Counter)
- **Description**: Total number of kernel cleanup operations
- **Labels**:
  - `success`: "True" or "False"

#### Event Propagator Metrics

External event propagation metrics for webhooks and integrations.

**`backendai_event_propagator_count`** (Gauge)
- **Description**: Current number of active event propagators
- **Labels**: None

**`backendai_event_propagator_alias_count`** (Gauge)
- **Description**: Current number of event propagator aliases
- **Labels**:
  - `domain`: Domain identifier for the alias
  - `alias_id`: Unique identifier for the propagator alias

**`backendai_event_propagator_registration_count`** (Counter)
- **Description**: Total number of event propagator registrations
- **Labels**: None

**`backendai_event_propagator_unregistration_count`** (Counter)
- **Description**: Total number of event propagator unregistrations
- **Labels**: None

#### Stage Metrics

Development and debugging metrics for tracking execution stages.

**`backendai_stage_count`** (Counter)
- **Description**: Count of stage occurrences for debugging and tracing
- **Labels**:
  - `stage`: Stage identifier
  - `upper_layer`: Calling layer or component

### Prometheus Query Examples

The following examples demonstrate common Prometheus queries for Manager metrics. Note that Counter metrics use the `_total` suffix and Histogram metrics use `_bucket`, `_sum`, `_count` suffixes in actual queries.

**Important Notes:**
- When using `increase()` or `rate()` functions, the time range must be at least 2-4x longer than your Prometheus scrape interval to get reliable data. If the time range is too short, metrics may not appear or show incomplete data.
- Default Prometheus scrape interval is typically 15s-30s
- **Time range selection trade-offs**:
  - Shorter ranges (e.g., `[1m]`): Detect changes faster with more granular data, but more sensitive to noise and short-term fluctuations
  - Longer ranges (e.g., `[5m]`): Smoother graphs with reduced noise, better for identifying trends, but slower to detect sudden changes
  - For real-time alerting: Use shorter ranges like `[1m]` or `[2m]`
  - For dashboards and trend analysis: Use longer ranges like `[5m]` or `[10m]`

#### API Request Rate

**API Request Rate by Endpoint**

Calculate the per-second rate of API requests grouped by endpoint and status. This shows how many requests per second each endpoint receives. Use this to identify high-traffic endpoints and monitor overall API load.

```promql
sum(rate(backendai_api_request_count_total{service_group="$service_groups"}[1m])) by (method, endpoint, status_code)
```

**Failed API Requests (5xx Errors)**

Monitor failed API requests (5xx errors) to identify service issues. This helps detect when the Manager is experiencing internal errors.

```promql
sum(rate(backendai_api_request_count_total{service_group="$service_groups", status_code=~"5.."}[5m])) by (endpoint)
```

#### API Request Duration

**P95 API Request Latency**

Calculate 95th percentile (P95) latency for API requests. This shows the response time that 95% of requests complete within. Use this to identify slow endpoints and set SLA targets.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_api_request_duration_sec_bucket{service_group="$service_groups"}[5m])) by (le, endpoint)
)
```

**Average API Request Duration**

Calculate average request duration per endpoint. This provides a simple overview of typical response times.

```promql
sum(rate(backendai_api_request_duration_sec_sum{service_group="$service_groups"}[5m])) by (endpoint)
/
sum(rate(backendai_api_request_duration_sec_count{service_group="$service_groups"}[5m])) by (endpoint)
```

#### GraphQL Query Performance

**GraphQL Query Rate by Operation**

Monitor GraphQL query rate by operation type and field. Use this to understand which GraphQL queries are most frequently used.

```promql
sum(rate(backendai_graphql_request_count_total{service_group="$service_groups"}[5m])) by (operation_type, field_name)
```

**Failed GraphQL Queries**

Track failed GraphQL queries with error details. This helps identify problematic queries and common error patterns.

```promql
sum(rate(backendai_graphql_request_count_total{service_group="$service_groups", success="False"}[5m])) by (field_name, error_detail)
```

#### Layer Operation Performance

**P95 Redis Operation Latency**

Monitor Redis operation latency (P95) by layer and operation. This helps identify slow Redis operations that may cause bottlenecks. Exclude broadcast/stream operations as they have different performance characteristics.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_layer_operation_duration_sec_bucket{domain="valkey", operation!~"receive_broadcast_message|read_consumer_group"}[5m])) by (le, layer, operation)
)
```

**P95 Database Repository Latency**

Monitor database repository operation latency (P95). Use this to identify slow database queries and optimize data access patterns.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_layer_operation_duration_sec_bucket{domain="repository"}[5m])) by (le, layer, operation)
)
```

**P95 Agent RPC Call Latency**

Monitor Agent RPC call latency (P95). This shows how long it takes to communicate with compute agents.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_layer_operation_duration_sec_bucket{domain="client", layer="agent_client"}[5m])) by (le, operation)
)
```

**Failed Layer Operations**

Track failed layer operations to identify integration issues. High error rates indicate problems with external dependencies or internal bugs.

```promql
sum(rate(backendai_layer_operation_count_total{success="False"}[5m])) by (domain, layer, operation)
```

#### Background Tasks

**Currently Running Background Tasks**

Monitor currently running background tasks. Gauge metric shows real-time count of active tasks. Use this to ensure background tasks are running and detect stuck tasks.

```promql
sum(backendai_bgtask_count) by (task_name)
```

**Background Task Completion Rate**

Track background task completion rate and success/failure status. This shows how frequently tasks complete and their success rate.

```promql
sum(rate(backendai_bgtask_done_count_total[5m])) by (task_name, status)
```

**Failed Background Tasks**

Monitor failed background tasks with error details. Use this to identify recurring task failures and debug issues.

```promql
sum(rate(backendai_bgtask_done_count_total{status="failure"}[5m])) by (task_name, error_detail)
```

#### Event Processing

**Event Processing Rate by Type**

Monitor event processing rate by event type. This shows how many events are being processed per second. Use this to understand event throughput and detect processing delays.

```promql
sum(rate(backendai_event_count_total[5m])) by (event_type)
```

**Event Processing Failures**

Track event processing failures by exception type. This helps identify problematic event handlers and common error patterns.

```promql
sum(rate(backendai_event_failure_count[5m])) by (event_type, exception)
```

**P95 Event Processing Duration**

Calculate P95 event processing duration. This shows how long it takes to process different types of events. Use this to identify slow event handlers that may cause delays.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_event_processing_time_sec_bucket[5m])) by (le, event_type)
)
```

#### Action Metrics

**Action Execution Rate**

Monitor action execution rate grouped by entity and operation type. This shows the rate of high-level business operations like session creation. Use this to understand system activity and user behavior patterns.

```promql
sum(rate(backendai_action_count_total[5m])) by (entity_type, operation_type, status)
```

**Failed Actions**

Track failed actions with detailed error information. This helps identify which operations are failing and why.

```promql
sum(rate(backendai_action_count_total{status="failure"}[5m])) by (entity_type, operation_type, error_detail)
```

**P95 Action Execution Duration**

Calculate P95 action execution duration. This shows how long key operations take to complete. Use this to set performance expectations and identify slow operations.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_action_duration_sec_bucket[5m])) by (le, entity_type, operation_type)
)
```

#### System Resources

Monitor active asyncio tasks in the event loop. High task counts may indicate resource leaks or excessive concurrency.

```promql
backendai_async_task_count
```

Monitor CPU usage percentage of the Manager process. Use this to detect CPU bottlenecks and capacity planning.

```promql
backendai_cpu_usage_percent
```

Monitor Resident Set Size (physical memory usage). This shows actual RAM usage by the Manager process.

```promql
backendai_memory_used_rss
```

Monitor Virtual Memory Size (total allocated memory). This includes swapped memory and memory-mapped files.

```promql
backendai_memory_used_vms
```

#### Session Sweeper

Monitor session cleanup operations by session status. This shows how many sessions are being cleaned up and success rate. Use this to ensure proper resource cleanup and identify cleanup issues.

```promql
sum(rate(backendai_sweep_session_count_total[5m])) by (status, success)
```

Monitor kernel cleanup operation rate. This tracks orphaned kernel cleanup operations.

```promql
sum(rate(backendai_sweep_kernel_count_total[5m])) by (success)
```

### Logs
- API request/response logs
- Session scheduling decision logs
- Resource allocation/release events
- Authentication and authorization events
- Background task execution logs
- Error and exception stack traces

## Communication Protocols

### Agent Communication
- **Protocol**: ZeroMQ RPC
- **Port**: 6011 (agent RPC server)
- **Main Operations**: Kernel lifecycle management, code execution, file operations, container statistics

### Storage Proxy Communication
- **Protocol**: HTTP
- **Port**: 6021 (client API), 6022 (manager API)
- **Main Operations**: VFolder management, file upload/download, file listing

### etcd Communication
- **Protocol**: gRPC (etcd v3 API)
- **Port**: 2379
- **Main Operations**: Configuration management, service discovery, watch notifications

## Development

See [README.md](./README.md) for development setup instructions.

## Manager Architecture Documentation

### Internal Architecture
- **[Sokovan Orchestration System](./sokovan/README.md)**: Session scheduling orchestrator with 3-tier architecture
  - [Scheduler](./sokovan/scheduler/README.md): Core scheduling engine for session placement and resource allocation
  - [Scheduling Controller](./sokovan/scheduling_controller/README.md): Validation and preparation logic for scheduling
  - [Deployment Controller](./sokovan/deployment/README.md): Deployment lifecycle management
- **[Services Layer](./services/README.md)**: Business logic patterns, design principles, and implementation guidelines
- **[Repositories Layer](./repositories/README.md)**: Data access patterns, query optimization, and transaction management
- **[Actions Layer](./actions/README.md)**: Permission validation, monitoring, and request handling

### Related Components
- [Agent Component](../agent/README.md): Kernel lifecycle management on compute nodes
- [Storage Proxy Component](../storage/README.md): Virtual folder and storage backend management
- [Webserver Component](../web/README.md): Web UI hosting and session management
- [Overall Architecture](../README.md): System-wide architecture and component interactions
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-manager',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.manager': (
            'VERSION',
            'py.typed',
        ),
        'ai.backend.manager.api': (
            'py.typed',
        ),
        'ai.backend.manager.models.alembic': (
            'revision_history/23.09.10.json',
            'revision_history/24.03.11.json',
            'revision_history/24.09.8.json',
            'script.py.mako',
        ),
        'ai.backend.manager.vendor': (
            'bai-icon.svg',
            'graphiql.min.css',
            'graphiql.min.js',
            'react-dom.production.min.js',
            'react.production.min.js',
            'spec-viewer.js',
        ),
    },
    'packages': (
        'ai.backend.manager',
        'ai.backend.manager.actions',
        'ai.backend.manager.actions.action',
        'ai.backend.manager.actions.monitors',
        'ai.backend.manager.actions.processor',
        'ai.backend.manager.actions.validator',
        'ai.backend.manager.actions.validators',
        'ai.backend.manager.actions.validators.rbac',
        'ai.backend.manager.api',
        'ai.backend.manager.api.gql',
        'ai.backend.manager.api.gql.data_loader',
        'ai.backend.manager.api.gql.model_deployment',
        'ai.backend.manager.api.gql.notification',
        'ai.backend.manager.api.gql.scaling_group',
        'ai.backend.manager.api.notification',
        'ai.backend.manager.api.vfolders',
        'ai.backend.manager.bgtask',
        'ai.backend.manager.bgtask.tasks',
        'ai.backend.manager.cli',
        'ai.backend.manager.client',
        'ai.backend.manager.client.artifact_registry',
        'ai.backend.manager.client.container_registry',
        'ai.backend.manager.clients',
        'ai.backend.manager.clients.agent',
        'ai.backend.manager.clients.appproxy',
        'ai.backend.manager.clients.storage_proxy',
        'ai.backend.manager.clients.valkey_client',
        'ai.backend.manager.clients.valkey_client.valkey_cache',
        'ai.backend.manager.config',
        'ai.backend.manager.config.loader',
        'ai.backend.manager.config.watchers',
        'ai.backend.manager.container_registry',
        'ai.backend.manager.data',
        'ai.backend.manager.data.agent',
        'ai.backend.manager.data.app_config',
        'ai.backend.manager.data.artifact',
        'ai.backend.manager.data.artifact_registries',
        'ai.backend.manager.data.association',
        'ai.backend.manager.data.auth',
        'ai.backend.manager.data.common',
        'ai.backend.manager.data.container_registry',
        'ai.backend.manager.data.deployment',
        'ai.backend.manager.data.domain',
        'ai.backend.manager.data.group',
        'ai.backend.manager.data.huggingface_registry',
        'ai.backend.manager.data.image',
        'ai.backend.manager.data.kernel',
        'ai.backend.manager.data.keypair',
        'ai.backend.manager.data.model_serving',
        'ai.backend.manager.data.notification',
        'ai.backend.manager.data.object_storage',
        'ai.backend.manager.data.permission',
        'ai.backend.manager.data.reservoir_registry',
        'ai.backend.manager.data.resource',
        'ai.backend.manager.data.resource_preset',
        'ai.backend.manager.data.scaling_group',
        'ai.backend.manager.data.session',
        'ai.backend.manager.data.storage_namespace',
        'ai.backend.manager.data.user',
        'ai.backend.manager.data.vfolder',
        'ai.backend.manager.data.vfs_storage',
        'ai.backend.manager.dependencies',
        'ai.backend.manager.dependencies.bootstrap',
        'ai.backend.manager.dependencies.components',
        'ai.backend.manager.dependencies.config',
        'ai.backend.manager.dependencies.infrastructure',
        'ai.backend.manager.dto',
        'ai.backend.manager.errors',
        'ai.backend.manager.event_dispatcher',
        'ai.backend.manager.event_dispatcher.handlers',
        'ai.backend.manager.events',
        'ai.backend.manager.events.hub',
        'ai.backend.manager.events.hub.propagators',
        'ai.backend.manager.health',
        'ai.backend.manager.metrics',
        'ai.backend.manager.models',
        'ai.backend.manager.models.alembic',
        'ai.backend.manager.models.alembic.versions',
        'ai.backend.manager.models.gql_models',
        'ai.backend.manager.models.gql_models.metric',
        'ai.backend.manager.models.hasher',
        'ai.backend.manager.models.minilang',
        'ai.backend.manager.models.rbac',
        'ai.backend.manager.models.rbac_models',
        'ai.backend.manager.models.rbac_models.migration',
        'ai.backend.manager.models.rbac_models.permission',
        'ai.backend.manager.network',
        'ai.backend.manager.notification',
        'ai.backend.manager.notification.channels',
        'ai.backend.manager.notification.channels.webhook',
        'ai.backend.manager.plugin',
        'ai.backend.manager.public_api',
        'ai.backend.manager.reporters',
        'ai.backend.manager.repositories',
        'ai.backend.manager.repositories.agent',
        'ai.backend.manager.repositories.agent.cache_source',
        'ai.backend.manager.repositories.agent.db_source',
        'ai.backend.manager.repositories.agent.stateful_source',
        'ai.backend.manager.repositories.app_config',
        'ai.backend.manager.repositories.app_config.cache_source',
        'ai.backend.manager.repositories.app_config.db_source',
        'ai.backend.manager.repositories.artifact',
        'ai.backend.manager.repositories.artifact.db_source',
        'ai.backend.manager.repositories.artifact_registry',
        'ai.backend.manager.repositories.artifact_registry.db_source',
        'ai.backend.manager.repositories.auth',
        'ai.backend.manager.repositories.auth.db_source',
        'ai.backend.manager.repositories.container_registry',
        'ai.backend.manager.repositories.deployment',
        'ai.backend.manager.repositories.deployment.db_source',
        'ai.backend.manager.repositories.deployment.storage_source',
        'ai.backend.manager.repositories.deployment.types',
        'ai.backend.manager.repositories.domain',
        'ai.backend.manager.repositories.group',
        'ai.backend.manager.repositories.huggingface_registry',
        'ai.backend.manager.repositories.huggingface_registry.db_source',
        'ai.backend.manager.repositories.image',
        'ai.backend.manager.repositories.image.db_source',
        'ai.backend.manager.repositories.image.stateful_source',
        'ai.backend.manager.repositories.keypair_resource_policy',
        'ai.backend.manager.repositories.keypair_resource_policy.db_source',
        'ai.backend.manager.repositories.metric',
        'ai.backend.manager.repositories.model_serving',
        'ai.backend.manager.repositories.notification',
        'ai.backend.manager.repositories.notification.db_source',
        'ai.backend.manager.repositories.object_storage',
        'ai.backend.manager.repositories.object_storage.db_source',
        'ai.backend.manager.repositories.permission_controller',
        'ai.backend.manager.repositories.project_resource_policy',
        'ai.backend.manager.repositories.reservoir_registry',
        'ai.backend.manager.repositories.reservoir_registry.db_source',
        'ai.backend.manager.repositories.resource_preset',
        'ai.backend.manager.repositories.resource_preset.cache_source',
        'ai.backend.manager.repositories.resource_preset.db_source',
        'ai.backend.manager.repositories.scaling_group',
        'ai.backend.manager.repositories.scaling_group.db_source',
        'ai.backend.manager.repositories.schedule',
        'ai.backend.manager.repositories.schedule.cache_source',
        'ai.backend.manager.repositories.schedule.db_source',
        'ai.backend.manager.repositories.schedule.types',
        'ai.backend.manager.repositories.scheduler',
        'ai.backend.manager.repositories.scheduler.cache_source',
        'ai.backend.manager.repositories.scheduler.db_source',
        'ai.backend.manager.repositories.scheduler.types',
        'ai.backend.manager.repositories.session',
        'ai.backend.manager.repositories.storage_namespace',
        'ai.backend.manager.repositories.storage_namespace.db_source',
        'ai.backend.manager.repositories.user',
        'ai.backend.manager.repositories.user_resource_policy',
        'ai.backend.manager.repositories.user_resource_policy.db_source',
        'ai.backend.manager.repositories.vfolder',
        'ai.backend.manager.repositories.vfs_storage',
        'ai.backend.manager.repositories.vfs_storage.db_source',
        'ai.backend.manager.scheduler',
        'ai.backend.manager.service',
        'ai.backend.manager.service.container_registry',
        'ai.backend.manager.services',
        'ai.backend.manager.services.agent',
        'ai.backend.manager.services.agent.actions',
        'ai.backend.manager.services.app_config',
        'ai.backend.manager.services.app_config.actions',
        'ai.backend.manager.services.artifact',
        'ai.backend.manager.services.artifact.actions',
        'ai.backend.manager.services.artifact_registry',
        'ai.backend.manager.services.artifact_registry.actions',
        'ai.backend.manager.services.artifact_registry.actions.common',
        'ai.backend.manager.services.artifact_registry.actions.huggingface',
        'ai.backend.manager.services.artifact_registry.actions.reservoir',
        'ai.backend.manager.services.artifact_revision',
        'ai.backend.manager.services.artifact_revision.actions',
        'ai.backend.manager.services.auth',
        'ai.backend.manager.services.auth.actions',
        'ai.backend.manager.services.container_registry',
        'ai.backend.manager.services.container_registry.actions',
        'ai.backend.manager.services.deployment',
        'ai.backend.manager.services.deployment.actions',
        'ai.backend.manager.services.deployment.actions.access_token',
        'ai.backend.manager.services.deployment.actions.auto_scaling_rule',
        'ai.backend.manager.services.deployment.actions.model_revision',
        'ai.backend.manager.services.domain',
        'ai.backend.manager.services.domain.actions',
        'ai.backend.manager.services.group',
        'ai.backend.manager.services.group.actions',
        'ai.backend.manager.services.image',
        'ai.backend.manager.services.image.actions',
        'ai.backend.manager.services.keypair_resource_policy',
        'ai.backend.manager.services.keypair_resource_policy.actions',
        'ai.backend.manager.services.metric',
        'ai.backend.manager.services.metric.actions',
        'ai.backend.manager.services.metric.processors',
        'ai.backend.manager.services.model_serving',
        'ai.backend.manager.services.model_serving.actions',
        'ai.backend.manager.services.model_serving.processors',
        'ai.backend.manager.services.model_serving.services',
        'ai.backend.manager.services.notification',
        'ai.backend.manager.services.notification.actions',
        'ai.backend.manager.services.object_storage',
        'ai.backend.manager.services.object_storage.actions',
        'ai.backend.manager.services.permission_contoller',
        'ai.backend.manager.services.permission_contoller.actions',
        'ai.backend.manager.services.project_resource_policy',
        'ai.backend.manager.services.project_resource_policy.actions',
        'ai.backend.manager.services.resource_preset',
        'ai.backend.manager.services.resource_preset.actions',
        'ai.backend.manager.services.scaling_group',
        'ai.backend.manager.services.scaling_group.actions',
        'ai.backend.manager.services.session',
        'ai.backend.manager.services.session.actions',
        'ai.backend.manager.services.storage_namespace',
        'ai.backend.manager.services.storage_namespace.actions',
        'ai.backend.manager.services.user',
        'ai.backend.manager.services.user.actions',
        'ai.backend.manager.services.user_resource_policy',
        'ai.backend.manager.services.user_resource_policy.actions',
        'ai.backend.manager.services.vfolder',
        'ai.backend.manager.services.vfolder.actions',
        'ai.backend.manager.services.vfolder.processors',
        'ai.backend.manager.services.vfolder.services',
        'ai.backend.manager.services.vfs_storage',
        'ai.backend.manager.services.vfs_storage.actions',
        'ai.backend.manager.sokovan',
        'ai.backend.manager.sokovan.deployment',
        'ai.backend.manager.sokovan.deployment.definition_generator',
        'ai.backend.manager.sokovan.deployment.handlers',
        'ai.backend.manager.sokovan.deployment.revision_generator',
        'ai.backend.manager.sokovan.deployment.route',
        'ai.backend.manager.sokovan.deployment.route.handlers',
        'ai.backend.manager.sokovan.scheduler',
        'ai.backend.manager.sokovan.scheduler.allocators',
        'ai.backend.manager.sokovan.scheduler.handlers',
        'ai.backend.manager.sokovan.scheduler.hooks',
        'ai.backend.manager.sokovan.scheduler.kernel',
        'ai.backend.manager.sokovan.scheduler.recorder',
        'ai.backend.manager.sokovan.scheduler.selectors',
        'ai.backend.manager.sokovan.scheduler.sequencers',
        'ai.backend.manager.sokovan.scheduler.validators',
        'ai.backend.manager.sokovan.scheduling_controller',
        'ai.backend.manager.sokovan.scheduling_controller.calculators',
        'ai.backend.manager.sokovan.scheduling_controller.preparers',
        'ai.backend.manager.sokovan.scheduling_controller.resolvers',
        'ai.backend.manager.sokovan.scheduling_controller.validators',
        'ai.backend.manager.sweeper',
        'ai.backend.manager.vendor',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': """25.18.0rc2
""",
    'zip_safe': False,
})
